part of charts;

class CartesianChartAnnotation {
  CartesianChartAnnotation(
      {this.child,
      CoordinateUnit coordinateUnit,
      AnnotationRegion region,
      ChartAlignment horizontalAlignment,
      ChartAlignment verticalAlignment,
      dynamic x,
      dynamic y,
      this.xAxisName,
      this.yAxisName})
      : coordinateUnit = coordinateUnit ?? CoordinateUnit.logicalPixel,
        region = region ?? AnnotationRegion.chart,
        x = x ?? 0,
        y = y ?? 0,
        horizontalAlignment = horizontalAlignment ?? ChartAlignment.center,
        verticalAlignment = verticalAlignment ?? ChartAlignment.center;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty Data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final Widget child;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60
  ///                 ),
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///```
  final CoordinateUnit coordinateUnit;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container( 
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final AnnotationRegion region;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final dynamic x;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final num y;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container( 
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60,
  ///                    xAxisName: 'Gold'
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final dynamic xAxisName;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60,
  ///                    yAxisName: 'Diamond'
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final dynamic yAxisName;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty data'), 
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60,
  ///                    HorizontalAlignment: HorizontalAlignment.near,
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final ChartAlignment horizontalAlignment;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container( 
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60,
  ///                    verticalAllignment: VerticalAlignment.bottom,
  ///                 ),
  ///               ),
  ///             ],
  ///        ));
  ///}
  ///```
  final ChartAlignment verticalAlignment;
}
