part of charts;

abstract class ChartAxis {
  ChartAxis(
      {this.name,
      double plotOffset,
      bool isVisible,
      AxisTitle title,
      AxisLine axisLine,
      ChartRangePadding rangePadding,
      int labelRotation,
      LabelPosition labelPosition,
      TickPosition tickPosition,
      MajorTickLines majorTickLines,
      MinorTickLines minorTickLines,
      ChartTextStyle labelStyle,
      AxisLabelIntersectAction labelIntersectAction,
      this.desiredIntervals,
      MajorGridLines majorGridLines,
      MinorGridLines minorGridLines,
      int maximumLabels,
      int minorTicksPerInterval,
      bool isInversed,
      bool opposedPosition,
      EdgeLabelPlacement edgeLabelPlacement,
      bool enableAutoIntervalOnZooming,
      double zoomFactor,
      double zoomPosition,
      InteractiveTooltip crosshairTooltip,
      this.interval})
      : isVisible = isVisible ?? true,
        crosshairTooltip = crosshairTooltip ?? InteractiveTooltip(),
        isInversed = isInversed ?? false,
        plotOffset = plotOffset ?? 0,
        opposedPosition = opposedPosition ?? false,
        rangePadding = rangePadding ?? ChartRangePadding.auto,
        labelRotation = labelRotation ?? 0,
        labelPosition = labelPosition ?? LabelPosition.outside,
        tickPosition = tickPosition ?? TickPosition.outside,
        labelIntersectAction =
            labelIntersectAction ?? AxisLabelIntersectAction.hide,
        minorTicksPerInterval = minorTicksPerInterval ?? 0,
        maximumLabels = maximumLabels ?? 3,
        labelStyle = _getTextStyle(
            textStyle: labelStyle,
            fontSize: 12.0,
            fontStyle: FontStyle.normal,
            fontWeight: FontWeight.normal,
            fontFamily: 'Normal'),
        title = title ?? AxisTitle(),
        axisLine = axisLine ?? AxisLine(),
        majorTickLines = majorTickLines ?? MajorTickLines(),
        minorTickLines = minorTickLines ?? MinorTickLines(),
        majorGridLines = majorGridLines ?? MajorGridLines(),
        minorGridLines = minorGridLines ?? MinorGridLines(),
        edgeLabelPlacement = edgeLabelPlacement ?? EdgeLabelPlacement.none,
        _bounds = Rect.fromLTWH(0, 0, 0, 0),
        zoomFactor = zoomFactor ?? 1,
        zoomPosition = zoomPosition ?? 0,
        enableAutoIntervalOnZooming = enableAutoIntervalOnZooming ?? true,    
        _visibleLabels = <AxisLabel>[] {
         _axisRenderer =  _ChartAxisRenderer();
        _maximumLabelSize = const Size(0, 0);
        _series = <CartesianSeries<dynamic, dynamic>>[];
        _name = name;
        _labelRotation = this.labelRotation;
        _zoomFactor = this.zoomFactor;
        _zoomPosition = this.zoomPosition;
        }

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, isVisible: false),
  ///        ));
  ///}
  ///```
  final bool isVisible;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, axisLine: AxisLine(width: 10)),
  ///        ));
  ///}
  ///```
  final AxisLine axisLine;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, majorTickLines: MajorTickLines(width: 0),),
  ///        ));
  ///}
  ///```
  /// This property is used to show or hide the major ticks of axis.
  final MajorTickLines majorTickLines;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, minorTickLines: MinorTickLines(width: 0),),
  ///        ));
  ///}
  ///```
  /// This property is used to show or hide the minor ticks of axis.
  final MinorTickLines minorTickLines;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, majorGridLines: MajorGridLines(width: 0)),
  ///        ));
  ///}
  ///```
  /// This property is used to show or hide the major ticks of axis.
  final MajorGridLines majorGridLines;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, minorGridLines: MinorGridLines(width: 0)),
  ///        ));
  ///}
  ///```
  /// This property is used to show or hide the minor grids of axis.
  final MinorGridLines minorGridLines;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(minimum: 0, maximum: 8, interval: 1, labelStyle: ChartTextStyle(color: Colors.black)),
  ///        ));
  ///}
  ///```
  /// This property is used to show or hide the axis labels.
  final ChartTextStyle labelStyle;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(title: AxisTitle( text: 'Year')),
  ///        ));
  ///}
  ///```
  /// This property specifies the axis title and also you can customize the font style.
  final AxisTitle title;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(rangePadding:  ChartRangePadding.round),
  ///        ));
  ///}
  ///```
  final ChartRangePadding rangePadding;

  /// Specifies the calculate intervals approximately equal to your specified interval.
  final int desiredIntervals;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(maximumLabels: 2),
  ///        ));
  ///}
  ///```
  final int maximumLabels;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(minorTicksPerInterval: 2),
  ///        ));
  ///}
  ///```
  final int minorTicksPerInterval;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(labelRotation: 20),
  ///        ));
  ///}
  ///```
  final int labelRotation;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(labelIntersectAction: AxisLabelIntersectAction.hide),
  ///        ));
  ///}
  ///```
  final AxisLabelIntersectAction labelIntersectAction;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(opposedPosition: true),
  ///        ));
  ///}
  ///```
  final bool opposedPosition;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(isInversed: true),
  ///        ));
  ///}
  ///```
  final bool isInversed;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(labelPosition: LabelPosition.inside),
  ///        ));
  ///}
  ///```
  final LabelPosition labelPosition;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(tickPosition: TickPosition.inside),
  ///        ));
  ///}
  ///```
  final TickPosition tickPosition;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(edgeLabelPlacement: EdgeLabelPlacement.hide),
  ///        ));
  ///}
  ///```
  final EdgeLabelPlacement edgeLabelPlacement;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(interval: 1),
  ///        ));
  ///}
  ///```
  final double interval;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: NumericAxis(plotOffset: 70),
  ///        ));
  ///}
  ///```
  final double plotOffset;

   _ChartAxisRenderer _axisRenderer;

  /// It holds the visible labels
  List<AxisLabel> _visibleLabels;

  /// It specifies the axis name
  final String name;

  /// It holds the size of larger label size
  Size _maximumLabelSize;

  /// It specifies axis orientations such as vertical, horizontal
  AxisOrientation _orientation;

  /// It specifies the visible range based on min, max values
  _VisibleRange _visibleRange;

  _VisibleRange _actualRange;

  /// It holds the chart series
  List<CartesianSeries<dynamic, dynamic>> _series;

  /// It holds the bounds of rectangle
  Rect _bounds;

  /// It is for zooming level
  final double zoomFactor;

  ///It is for zooming position
  final double zoomPosition;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(enableAutoIntervalOnZooming: true),
  ///        ));
  ///}
  ///```
  final bool enableAutoIntervalOnZooming;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           primaryXAxis: DateTimeAxis(crosshairTooltip: InteractiveTooltip(enable: true)),
  ///        ));
  ///}
  ///```
  final InteractiveTooltip crosshairTooltip;

  bool _isInsideTickPosition;

  double _totalSize;

  ///Internal variable
  double _previousZoomFactor;

  ///Internal variable
  double _previousZoomPosition;

  ///Internal variables

  String _name;

  int _labelRotation;

  double _zoomFactor;

  double _zoomPosition;


  ///Checking the axis label collision
  bool _isCollide = false;

  SfCartesianChart _chart;

  void calculateRange();

  num calculateInterval(_VisibleRange range, Size availableSize);

  void applyRangePadding(_VisibleRange range, num interval);

  void calculateVisibleRange(Size availableSize);

  void generateVisibleLabels();

  /// Calculate the interval based min and max values in axis
  num _calculateNumericNiceInterval(ChartAxis axis, num delta, Size size) {
    final dynamic intervalDivisions = <dynamic>[10, 5, 2, 1];
    num niceInterval;

    /// Get the desired interval if desired interval not specified
    final num actualDesiredIntervalCount =
        _calculateDesiredIntervalCount(size, axis);
    niceInterval = delta / actualDesiredIntervalCount;
    if (axis.desiredIntervals != null) {
      return niceInterval;
    }

    /// Get the minimum interval
    final num minimumInterval =
        niceInterval == 0 ? 0 : math.pow(10, _logBase(niceInterval, 10).floor());
    for (int i = 0; i < intervalDivisions.length; i++) {
      final num interval = intervalDivisions[i];
      final num currentInterval = minimumInterval * interval;
      if (actualDesiredIntervalCount < (delta / currentInterval)) {
        break;
      }
      niceInterval = currentInterval;
    }
    return niceInterval;
  }

  /// Calculate the axis interval for numeric axis
  num _calculateDesiredIntervalCount(Size availableSize, ChartAxis axis) {
    final dynamic size = axis._orientation == AxisOrientation.horizontal
        ? availableSize.width
        : availableSize.height;
    if (axis.desiredIntervals == null) {
      num desiredIntervalCount =
          (axis._orientation == AxisOrientation.horizontal ? 0.533 : 1) *
              axis.maximumLabels;
      desiredIntervalCount = math.max(size * (desiredIntervalCount / 100), 1);
      return desiredIntervalCount;
    } else {
      return axis.desiredIntervals;
    }
  }

  /// Get the range padding of an axis
  ChartRangePadding _calculateRangePadding(ChartAxis axis, SfCartesianChart chart) {
    ChartRangePadding padding;
    if (axis.rangePadding != ChartRangePadding.auto) {
      padding = axis.rangePadding;
    } else if (axis.rangePadding == ChartRangePadding.auto &&
        axis._orientation != null) {
      switch (axis._orientation) {
        case AxisOrientation.horizontal:
          padding = (chart._requireInvertedAxis)
              ? ChartRangePadding.normal
              : ChartRangePadding.none;
          break;
        case AxisOrientation.vertical:
          padding = (!chart._requireInvertedAxis)
              ? ChartRangePadding.normal
              : ChartRangePadding.none;
          break;
      }
    }
    return padding;
  }

  /// Calculate the maximum lable's size
  void _calculateMaximumLabelSize(ChartAxis axis, SfCartesianChart chart) {
    AxisLabelIntersectAction action;
    int angle = axis._labelRotation;
    AxisLabel label;
    Size currentLabelSize;
    num maximumLabelHeight = 0.0,
        maximumLabelWidth = 0.0,
        labelMaximumWidth,
        pointX,
        height;
    action = axis.labelIntersectAction;
    labelMaximumWidth =
        chart._chartAxis._axisClipRect.width / axis._visibleLabels.length;

    if (axis._orientation == AxisOrientation.horizontal &&
        axis.labelIntersectAction != AxisLabelIntersectAction.none &&
        axis._visibleLabels.length > 1) {
      final Rect axisBounds = chart._chartAxis._axisClipRect;
      AxisLabel label1;
      num pointX1;
      for (int i = 0; i < axis._visibleLabels.length - 1; i++) {
        label = axis._visibleLabels[i];
        pointX = (_valueToCoefficient(label.value, axis) * axisBounds.width) +
            axisBounds.left;
        pointX -= label.labelSize.width / 2;
        pointX = (i == 0 &&
                axis.edgeLabelPlacement == EdgeLabelPlacement.shift &&
                ((pointX < axisBounds.left && !axis.isInversed) ||
                    (pointX + label.labelSize.width > axisBounds.right &&
                        axis.isInversed)))
            ? (axis.isInversed
                ? axisBounds.left + axisBounds.width - label.labelSize.width
                : axisBounds.left)
            : ((i == axis._visibleLabels.length - 1 &&
                    axis.edgeLabelPlacement == EdgeLabelPlacement.shift &&
                    ((((pointX + label.labelSize.width) > axisBounds.right) &&
                            !axis.isInversed) ||
                        (pointX < axisBounds.left && axis.isInversed)))
                ? (axis.isInversed
                    ? axisBounds.left
                    : axisBounds.left +
                        axisBounds.width -
                        label.labelSize.width)
                : pointX);

        label1 = axis._visibleLabels[i + 1];
        pointX1 = (_valueToCoefficient(label1.value, axis) *
                chart._chartAxis._axisClipRect.width) +
            chart._chartAxis._axisClipRect.left;
        pointX1 -= label1.labelSize.width / 2;

        if ((((pointX + label.labelSize.width) > pointX1) &&
                !axis.isInversed) ||
            (((pointX - label.labelSize.width) < pointX1) && axis.isInversed)) {
          _isCollide = true;
          break;
        }
      }
    }

    for (int i = 0; i < axis._visibleLabels.length; i++) {
      label = axis._visibleLabels[i];
      if (label.labelSize.width > maximumLabelWidth) {
        maximumLabelWidth = label.labelSize.width;
      }
      if (label.labelSize.height > maximumLabelHeight) {
        maximumLabelHeight = label.labelSize.height;
      }

      if (axis._orientation == AxisOrientation.horizontal) {
        pointX = (_valueToCoefficient(label.value, axis) *
                chart._chartAxis._axisClipRect.width) +
            chart._chartAxis._axisClipRect.left;
        pointX -= label.labelSize.width / 2;

        /// Based on below options, perform label intersection
        if (_isCollide) {
          switch (action) {
            case AxisLabelIntersectAction.multipleRows:
              if (i > 0) {
                height = _findMultiRows(i, pointX, label, axis, chart);
                if (height > maximumLabelHeight) {
                  maximumLabelHeight = height;
                }
              }
              break;
            case AxisLabelIntersectAction.rotate45:
            case AxisLabelIntersectAction.rotate90:
              // if (i > 0 &&
              //     (!axis.isInversed
              //         ? pointX <= previousEnd
              //         : pointX + label.labelSize.width >= previousEnd)) {
              if (i > 0) {
                angle = action == AxisLabelIntersectAction.rotate45 ? 45 : 90;
                axis._labelRotation = angle;
                currentLabelSize =
                    _measureText(label.text, axis.labelStyle, angle);
                if (currentLabelSize.height > maximumLabelHeight)
                  maximumLabelHeight = currentLabelSize.height;
                if (currentLabelSize.width > maximumLabelWidth)
                  maximumLabelWidth = currentLabelSize.width;
              }
              break;
            case AxisLabelIntersectAction.wrap:
              label._labelCollection =
                  _textWrap(label.text, labelMaximumWidth, axis);
              if (label._labelCollection.isNotEmpty)
                label.text = label._labelCollection[0];
              height = label.labelSize.height * label._labelCollection.length;
              if (height > maximumLabelHeight) {
                maximumLabelHeight = height;
              }
              break;
            default:
              break;
          }
        }
      }
    }
    axis._maximumLabelSize = Size(maximumLabelWidth, maximumLabelHeight);
  }

  num _findMultiRows(num length, num currentX, AxisLabel currentLabel,
      ChartAxis axis, SfCartesianChart chart) {
    AxisLabel label;
    num pointX;
    final dynamic labelIndex = <dynamic>[];
    bool isMultiRows;
    for (int i = length - 1; i >= 0; i--) {
      label = axis._visibleLabels[i];
      pointX = (_valueToCoefficient(label.value, axis) *
              chart._chartAxis._axisClipRect.width) +
          chart._chartAxis._axisClipRect.left;
      isMultiRows = !axis.isInversed
          ? currentX < (pointX + label.labelSize.width / 2)
          : currentX + currentLabel.labelSize.width >
              (pointX - label.labelSize.width / 2);
      if (isMultiRows) {
        labelIndex.add(label._index);
        currentLabel._index = (currentLabel._index > label._index)
            ? currentLabel._index
            : label._index + 1;
      } else {
        currentLabel._index =
            labelIndex.contains(label._index) ? currentLabel._index : label._index;
      }
    }
    return currentLabel.labelSize.height * currentLabel._index;
  }

  List<String> _textWrap(String currentLabel, num maximumWidth, ChartAxis axis) {
    final List<String> textCollection = currentLabel.split(RegExp(' '));
    final List<String> labelCollection = <String>[];
    String text;
    for (int i = 0; i < textCollection.length; i++) {
      text = textCollection[i];
      if (_measureText(text, axis.labelStyle, axis._labelRotation).width <
          maximumWidth) {
        labelCollection.add(text);
      } else {
        labelCollection
            .add(_textTrim(text, maximumWidth, axis.labelStyle, axis));
      }
    }
    return labelCollection;
  }

  String _textTrim(String text, num maximumWidth, ChartTextStyle labelStyle,
      ChartAxis axis) {
    String label = text;
    num size = _measureText(text, axis.labelStyle, axis._labelRotation).width;
    if (size > maximumWidth) {
      final int textLength = text.length;
      for (int i = textLength - 1; i >= 0; --i) {
        label = text.substring(0, i) + '...';
        size = _measureText(label, labelStyle, axis._labelRotation).width;
        if (size <= maximumWidth) {
          return label;
        }
      }
    }
    return label;
  }

  ///Below method is for changing range while zooming
  void _calculateZoomRange(ChartAxis axis, Size axisSize) {
    final _VisibleRange baseRange = axis._visibleRange;
    num start, end;
    start =
        axis._visibleRange.minimum + axis._zoomPosition * axis._visibleRange.delta;
    end = start + axis._zoomFactor * axis._visibleRange.delta;

    if (start < baseRange.minimum) {
      end = end + (baseRange.minimum - start);
      start = baseRange.minimum;
    }
    if (end > baseRange.maximum) {
      start = start - (end - baseRange.maximum);
      end = baseRange.maximum;
    }
    axis._visibleRange.minimum = start;
    axis._visibleRange.maximum = end;
    axis._visibleRange.delta = end - start;
  }

  void _checkWithZoomState(ChartAxis axis, List<ChartAxis> axisStates) {
    for (int i = 0; i < axisStates.length; i++) {
      final ChartAxis zoomedAxis = axisStates[i];
      if (zoomedAxis._name == axis._name) {
        axis._zoomFactor = zoomedAxis._zoomFactor;
        axis._zoomPosition = zoomedAxis._zoomPosition;
        break;
      }
    }
  }
}

class AxisLabel {
  AxisLabel(this.labelStyle, this.labelSize, this.text, this.value);

  /// It specifies the textstyle of label
  ChartTextStyle labelStyle;

  /// It holds size of label
  Size labelSize;

  /// It holds text of label to be print
  String text;

  num value;

  List<String> _labelCollection;

  int _index = 1;

  ///Stores the location of an label
  Rect _labelRegion;
}

class MajorTickLines {
  MajorTickLines({this.size = 5, this.width = 1, this.color});

  /// Specifies the height in pixels.
  final double size;

  /// Specifies the width in pixels.
  final double width;

  /// Specifies the color for ticks.
  final Color color;
}

class MinorTickLines {
  MinorTickLines({this.size = 3, this.width = 0.7, this.color});

  /// Specifies the height in pixels.
  final double size;

  /// Specifies the width in pixels.
  final double width;

  /// Specifies the color for ticks.
  final Color color;
}

class MajorGridLines {
  MajorGridLines({this.width = 0.7, this.color, this.dashArray});

  ///Specifies the dash lines.
  final List<double> dashArray;

  /// Specifies the width in pixels.
  final double width;

  /// Specifies the color which is applied in grid lines.
  final Color color;
}

class MinorGridLines {
  MinorGridLines({this.width = 0.5, this.color, this.dashArray});

  /// Specifies the dash lines.
  final List<double> dashArray;

  /// Specifies the width in pixels.
  final double width;

  /// Specifies the color which is applied in minor grid lines.
  final Color color;
}

class AxisTitle {
  AxisTitle(
      {this.text,
      ChartTextStyle textStyle,
      this.alignment = ChartAlignment.center})
      : textStyle = _getTextStyle(
            textStyle: textStyle,
            fontFamily: 'Segoe UI',
            fontSize: 15.0,
            fontStyle: FontStyle.normal,
            fontWeight: FontWeight.normal);

  ///Specifies the title of an axis.
  final String text;

  /// Specifies the title style of an axis.
  final ChartTextStyle textStyle;

  /// It is defined that align the title based on mention options.
  final ChartAlignment alignment;
}

class AxisLine {
  AxisLine({this.color, this.dashArray, this.width = 1});

  ///Specifies the width in pixels.
  final double width;

  /// Specifies the color which is applied in axis line.
  final Color color;

  /// Specifies the dash lines for axis line
  final List<double> dashArray;
}

///calculate visible range based on min, max values
class _VisibleRange {
  _VisibleRange(dynamic min, dynamic max) {
    if (min < max) {
      minimum = min;
      maximum = max;
    } else {
      minimum = max;
      maximum = min;
    }
  }

  /// specifies minimum range
  dynamic minimum;

  /// Specifies maximum range
  dynamic maximum;

  /// Specifies range interval
  dynamic interval;

  num delta;

  // num get median => minimum + ((maximum - minimum) / 2);
}
