part of charts;

typedef ChartTooltipCallback = TooltipArgs Function(TooltipArgs tooltipArgs);
typedef ChartActualRangeChangedCallback = ActualRangeChangedArgs Function(
    ActualRangeChangedArgs rangeChangedArgs);
typedef ChartAxisLabelRenderCallback = AxisLabelRenderArgs Function(
    AxisLabelRenderArgs axisLabelRenderArgs);
typedef ChartDatalabelRenderCallback = DataLabelRenderArgs Function(
    DataLabelRenderArgs dataLabelArgs);
typedef ChartLegendRenderCallback = LegendRenderArgs Function(
    LegendRenderArgs legendRenderArgs);
typedef ChartTrackballCallback = TrackballArgs Function(
    TrackballArgs trackballArgs);
typedef ChartCrosshairCallback = CrosshairRenderArgs Function(
    CrosshairRenderArgs crosshairArgs);
typedef ChartZoomingCallback = ZoomPanArgs Function(ZoomPanArgs zoomingArgs);
typedef ChartPointTapCallback = PointTapArgs Function(
    PointTapArgs pointTapArgs);
typedef ChartAxisLabelTapCallback = AxisLabelTapArgs Function(
    AxisLabelTapArgs axisLabelTapArgs);
typedef ChartLegendTapCallback = LegendTapArgs Function(
    LegendTapArgs legendTapArgs);
typedef ChartSelectionCallback = SelectionArgs Function(
    SelectionArgs selectionArgs);

///```dart
///Widget build(BuildContext context) {
///  return Center(
///    child:SfCartesianChart(
///      title: ChartTitle(text: 'Flutter Chart'),
///     legend: Legend(isVisible: true),
///     series: getDefaultData(),
///     tooltipBehavior: TooltipBehavior(enable: true),
///    )
/// );
///}
///static List<LineSeries<SalesData, num>> getDefaultData() {
///    final dynamic chartData = <SalesData>[
///      SalesData(DateTime(2005, 0, 1), 'India', 1.5, 21, 28, 680, 760),
///      SalesData(DateTime(2006, 0, 1), 'China', 2.2, 24, 44, 550, 880),
///      SalesData(DateTime(2007, 0, 1), 'USA', 3.32, 36, 48, 440, 788),
///      SalesData(DateTime(2008, 0, 1), 'Japan', 4.56, 38, 50, 350, 560),
///      SalesData(DateTime(2009, 0, 1), 'Russia', 5.87, 54, 66, 444, 566),
///      SalesData(DateTime(2010, 0, 1), 'France', 6.8, 57, 78, 780, 650),
///     SalesData(DateTime(2011, 0, 1), 'Germany', 8.5, 70, 84, 450, 800)
///    ];
///   return <LineSeries<SalesData, num>>[
///      LineSeries<SalesData, num>(
///          enableToolTip: isTooltipVisible,
///          dataSource: chartData,
///          xValueMapper: (SalesData sales, _) => sales.numeric,
///          yValueMapper: (SalesData sales, _) => sales.sales1,
///          width: lineWidth ?? 2,
///          enableAnimation: false,
///         markerSettings: MarkerSettings(
///              isVisible: isMarkerVisible,
///              height: markerWidth ?? 4,
///              width: markerHeight ?? 4,
///              shape: DataMarkerType.Circle,
///              borderWidth: 3,
///              borderColor: Colors.red),
///          dataLabelSettings: DataLabelSettings(
///              visible: isDataLabelVisible, position: CartesianLabelPosition.Auto)),
///      LineSeries<SalesData, num>(
///          enableToolTip: isTooltipVisible,
///          dataSource: chartData,
///          enableAnimation: false,
///          width: lineWidth ?? 2,
///          xValueMapper: (SalesData sales, _) => sales.numeric,
///          yValueMapper: (SalesData sales, _) => sales.sales2,
///          markerSettings: MarkerSettings(
///              isVisible: isMarkerVisible,
///              height: markerWidth ?? 4,
///              width: markerHeight ?? 4,
///              shape: DataMarkerType.Circle,
///              borderWidth: 3,
///              borderColor: Colors.black),
///          dataLabelSettings: DataLabelSettings(
///              isVisible: isDataLabelVisible, position: CartesianLabelPosition.Auto))
///    ];
///  }
///  ```
///  {@end-tool}
///
// ignore: must_be_immutable
class SfCartesianChart extends StatefulWidget {
  SfCartesianChart(
      {Key key,
      this.backgroundColor = Colors.transparent,
      this.enableSideBySideSeriesPlacement = true,
      this.borderColor = Colors.transparent,
      this.borderWidth = 0,
      this.plotAreaBackgroundColor,
      this.plotAreaBorderColor = const Color.fromRGBO(219, 219, 219, 1),
      this.plotAreaBorderWidth = 0.7,
      this.backgroundImage,
      this.onTooltipRender,
      this.onActualRangeChanged,
      this.onAxisLabelRender,
      this.onDataLabelRender,
      this.onLegendItemRender,
      this.onTrackballPositionChanging,
      this.onCrosshairPositionChanging,
      this.onZoomStart,
      this.onZoomEnd,
      this.onZoomReset,
      this.onPointTapped,
      this.onAxisLabelTapped,
      this.onLegendTapped,
      this.onSelectionChanged,
      this.isTransposed = false,
      this.annotations,
      this.palette = const <Color>[
        Color.fromRGBO(53,92,125,1),
        Color.fromRGBO(192,108,132,1),
        Color.fromRGBO(246,114,128,1),
        Color.fromRGBO(248,177,149,1),
        Color.fromRGBO(116,180,155,1),
        Color.fromRGBO(0,168,181,1),
        Color.fromRGBO(73,76,162,1),
        Color.fromRGBO(255,205,96,1),
        Color.fromRGBO(255,240,219,1),
        Color.fromRGBO(238,238,238,1)
      ],
      ChartAxis primaryXAxis,
      ChartAxis primaryYAxis,
      EdgeInsets margin,
      TooltipBehavior tooltipBehavior,
      ZoomPanBehavior zoomPanBehavior,
      Legend legend,
      SelectionType selectionType,
      ActivationMode selectionGesture,
      bool enableMultiSelection,
      CrosshairBehavior crosshairBehavior,
      TrackballBehavior trackballBehavior,
      dynamic series,
      List<IndexesModel> initialSelectedDataIndexes,
      ChartTitle title,
      List<ChartAxis> axes})
      : primaryXAxis = primaryXAxis ?? NumericAxis(),
        primaryYAxis = primaryYAxis ?? NumericAxis(),
        title = title ?? ChartTitle(),
        axes = axes ?? <ChartAxis>[],
        series = series ?? <ChartSeries<dynamic, dynamic>>[],
        initialSelectedDataIndexes =
            initialSelectedDataIndexes ?? <IndexesModel>[],
        margin = margin ?? const EdgeInsets.all(10),
        zoomPanBehavior = zoomPanBehavior ?? ZoomPanBehavior(),
        tooltipBehavior = tooltipBehavior ?? TooltipBehavior(),
        crosshairBehavior =
            crosshairBehavior ?? CrosshairBehavior(),
        trackballBehavior =
            trackballBehavior ?? TrackballBehavior(),
        legend = legend ?? Legend(),
        selectionType = selectionType ?? SelectionType.point,
        selectionGesture = selectionGesture ?? ActivationMode.singleTap,
        enableMultiSelection = enableMultiSelection ?? false,
        super(key: key) {
    _chartAxis = _ChartAxis();
    _chartSeries = _ChartSeries();
    _chartLegend = _ChartLegend();
    _chartTheme = _ChartTheme();
  }
  

  /// The title to use for this chart.
  ///
  /// The default value is [ChartTitle].
  /// for example:
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            title: ChartTitle(
  ///                    text: 'Area with animation',
  ///                    alignment: ChartAlignment.center,
  ///                    backgroundColor: Colors.white,
  ///                    borderColor: Colors.transparent,
  ///                    borderWidth: 0)
  ///        ));
  ///}
  ///```
  final ChartTitle title;

  /// Options for customizing the legend in chart.
  /// If legend is enabled, the legend bounds will be calculate initially before axis calculate.
  ///
  /// The default value is [Legend].
  /// for example:
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            legend: Legend(isVisible: true),
  ///        ));
  ///}
  ///```
  final Legend legend;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            backgroundColor: Colors.blue
  ///        ));
  ///}
  ///```
  /// The default value is transparent.
  final Color backgroundColor;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            borderColor: Colors.red
  ///        ));
  ///}
  ///```
  /// The default value is transparent.
  final Color borderColor;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            borderColor: Colors.red,
  ///            borderWidth: 2
  ///        ));
  ///}
  ///```
  /// The default value is 0.
  final double borderWidth;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            plotAreaBackgroundColor: Colors.red,
  ///        ));
  ///}
  ///```
  /// The default value is null.
  final Color plotAreaBackgroundColor;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            plotAreaBorderColor: Colors.red,
  ///        ));
  ///}
  ///```
  /// The default value is grey.
  final Color plotAreaBorderColor;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            plotAreaBorderColor: Colors.red,
  ///            plotAreaBorderWidth: 2
  ///        ));
  ///}
  ///```
  /// The default value is 0.
  final double plotAreaBorderWidth;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            primaryXAxis: DateTimeAxis(interval: 1)
  ///        ));
  ///}
  ///```
  /// The default value is [NumericAxis].
  final ChartAxis primaryXAxis;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            primaryYAxis: NumericAxis(isinversed: false)
  ///        ));
  ///}
  ///```
  /// The default value is [NumericAxis].
  final ChartAxis primaryYAxis;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            margin: const EdgeInsets.all(2),
  ///            borderColor: Colors.blue
  ///        ));
  ///}
  ///```
  /// The default value is [EdgeInsets].
  final EdgeInsets margin;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///             axes: <ChartAxis>[
  ///                NumericAxis( majorGridLines: MajorGridLines(color: Colors.transparent))
  ///              ]
  ///        ));
  ///}
  ///```
  final List<ChartAxis> axes;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           enableSideBySideSeriesPlacement: false
  ///        ));
  ///}
  ///```
  /// The default value is true.
  final bool enableSideBySideSeriesPlacement;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            tooltipBehavior: TooltipBehavior(enable: true)
  ///            onTooltipRender: (TooltipArgs args) => tool(args)
  ///        ));
  ///}
  ///dynamic tool(TooltipArgs args) {
  ///   args.locationX = 30;
  ///}
  ///```
  final ChartTooltipCallback onTooltipRender;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onActualRangeChanged: (ActualRangeChangedArgs args) => range(args)
  ///        ));
  ///}
  ///dynamic range(ActualRangeChangedArgs args) {
  ///   print(args.visibleMin);
  ///}
  ///```
  final ChartActualRangeChangedCallback onActualRangeChanged;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onAxisLabelRender: (AxisLabelRenderArgs args) => axis(args),
  ///        ));
  ///}
  ///dynamic axis(AxisLabelRenderArgs args) {
  ///   args.text = 'axis Label';
  ///}
  ///```
  final ChartAxisLabelRenderCallback onAxisLabelRender;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onDataLabelRender: (DataLabelRenderArgs args) => dataLabel(args),
  ///        ));
  ///}
  ///dynamic dataLabel(DataLabelRenderArgs args) {
  ///   args.text = 'data Label';
  ///}
  ///```
  final ChartDatalabelRenderCallback onDataLabelRender;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            legend: Legend(isVisible: true),
  ///            onLegendItemRender: (LegendRendererArgs args) => legend(args)
  ///        ));
  ///}
  ///dynamic legend(LegendRendererArgs args) {
  ///   args.seriesIndex = 2;
  ///}
  ///```
  final ChartLegendRenderCallback onLegendItemRender;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            trackballBehavior: TrackballBehavior(enable: true),
  ///            onTrackballPositionChanging: (TrackballArgs args) => trackball(args)
  ///        ));
  ///}
  ///dynamic trackball(TrackballArgs args) {
  ///    args.chartPointInfo = ChartPointInfo();
  ///}
  ///```
  final ChartTrackballCallback onTrackballPositionChanging;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            crosshairBehavior: CrosshairBehavior(enable: true),
  ///            onCrosshairPositionChanging: (CrosshairRenderArgs args) => crosshair(args)
  ///        ));
  ///}
  ///dynamic crosshair(CrosshairRenderArgs args) {
  ///    args.text = 'crosshair';
  ///}
  ///```
  final ChartCrosshairCallback onCrosshairPositionChanging;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            zoomPanBehavior: ZoomPanBehavior(enableSelectionZooming: true),
  ///            onZoomStart: (ZoomingArgs args) => zoom(args),
  ///        ));
  ///}
  ///dynamic zoom(ZoomingArgs args) {
  ///    args.currentZoomFactor = 0.2;
  ///}
  ///```
  final ChartZoomingCallback onZoomStart;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            zoomPanBehavior: ZoomPanBehavior(enableSelectionZooming: true),
  ///            onZoomEnd: (ZoomingArgs args) => zoom(args),
  ///        ));
  ///}
  ///dynamic zoom(ZoomingArgs args) {
  ///    args.currentZoomPosition = 1;
  ///}
  ///```
  final ChartZoomingCallback onZoomEnd;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            zoomPanBehavior: ZoomPanBehavior(enableSelectionZooming: true),
  ///            onZoomReset: (ZoomingArgs args) => zoom(args),
  ///        ));
  ///}
  ///dynamic zoom(ZoomingArgs args) {
  ///    args.currentZoomPosition = 1;
  ///}
  ///```
  final ChartZoomingCallback onZoomReset;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onPointTapped: (PointTapArgs args) => point(args),
  ///        ));
  ///}
  ///dynamic point(PointTapArgs args) {
  ///   print(args.seriesIndex);
  ///}
  ///```
  final ChartPointTapCallback onPointTapped;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onAxisLabelTapped: (AxisLabelTapArgs args) => axis(args),
  ///        ));
  ///}
  ///dynamic axis(AxisLabelTapArgs args) {
  ///   print(args.text);
  ///}
  ///```
  final ChartAxisLabelTapCallback onAxisLabelTapped;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onLegendTapped: (LegendTapArgs args) => legend(args),
  ///        ));
  ///}
  ///dynamic legend(LegendTapArgs args) {
  ///   print(args.pointIndex);
  ///}
  ///```
  final ChartLegendTapCallback onLegendTapped;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            onSelectionChanged: (SelectionArgs args) => print(args.selectedColor),
  ///        ));
  ///}
  final ChartSelectionCallback onSelectionChanged;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            tooltipBehavior: TooltipBehavior(enable: true)
  ///        ));
  ///}
  ///```
  /// The default value is [TooltipBehavior].
  final TooltipBehavior tooltipBehavior;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            crosshairBehavior: CrosshairBehavior(enable: true),
  ///        ));
  ///}
  ///```
  /// The default value is [CrosshairBehavior].
  final CrosshairBehavior crosshairBehavior;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            trackballBehavior: TrackballBehavior(enable: true),
  ///        ));
  ///}
  ///```
  /// The default value is [TrackballBehavior].
  final TrackballBehavior trackballBehavior;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            zoomPanBehavior: ZoomPanBehavior( enablePanning: true),
  ///        ));
  ///}
  ///```
  /// The default value is [ZoomPanBehaviour].
  final ZoomPanBehavior zoomPanBehavior;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionType: SelectionType.point,
  ///        ));
  ///}
  ///```
  /// The default value is [SelectionType.point].
  final SelectionType selectionType;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            annotations: <CartesianChartAnnotation>[
  ///                CartesianChartAnnotation(
  ///                    child: Container(
  ///                    child: const Text('Empty data'),
  ///                    coordinateUnit: CoordinateUnit.point,
  ///                    region: AnnotationRegion.chartArea,
  ///                    x: 3.5,
  ///                    y: 60
  ///                 ),
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///```
  /// The default value is [CartesianChartAnnotation].
  final List<CartesianChartAnnotation> annotations;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            enableMultiSelection: true,
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  /// The default value is false.
  final bool enableMultiSelection;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionGesture: ActivationMode.doubleTap,
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  /// The default value is SingleTap.
  final ActivationMode selectionGesture;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            backgroundImage: const AssetImage('images/bike.png'),
  ///        ));
  ///}
  ///```
  final ImageProvider backgroundImage;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           initialSelectedDataIndexes: <IndexesModel>[IndexesModel(2,0)],
  ///           series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final List<IndexesModel> initialSelectedDataIndexes;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            isTransposed: true,
  ///        ));
  ///}
  ///```
  final bool isTransposed;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           series: <AreaSeries<SalesData, num>>[
  ///                AreaSeries<SalesData, num>(
  ///                    dataSource: chartData,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final List<ChartSeries<dynamic, dynamic>> series;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            palette: <Color>[Colors.red, Colors.green]
  ///        ));
  ///}
  ///```
  final List<Color> palette;

  /// Setting series animation duration factor
  final double _seriesDurationFactor = 0.85;

  /// Holds the information of chart state class
  _SfCartesianChartState _chartState;

  /// Whether to check chart axis is inverted or not
  bool _requireInvertedAxis;

  /// Holds the information of AxisBase class
  _ChartAxis _chartAxis;

  /// Holds the information of SeriesBase class
  _ChartSeries _chartSeries;

  /// Holds the information of LegendBase class
  _ChartLegend _chartLegend;

  /// Holds the information of ChartTheme class
  _ChartTheme _chartTheme;

  @override
  State<StatefulWidget> createState() {
    return _SfCartesianChartState();
  }
}

class _SfCartesianChartState extends State<SfCartesianChart>
    with TickerProviderStateMixin {
  AnimationController animationController; // Animation controller for series
  List<AnimationController>
      controllerList; // Holds the animation controller list for all series
  ValueNotifier<int> zoomRepaintNotifier; // Repaint notifier for zoom container
  ValueNotifier<int> axisRepaintNotifier; // Repaint notifier for axis container
  ValueNotifier<int>
      seriesRepaintNotifier; // Repaint notifier for series container
  ValueNotifier<int>
      tooltipRepaintNotifier; // Repaint notifier for tooltip container
  ValueNotifier<int>
      trackballRepaintNotifier; // Repaint notifier for trackball container
  ValueNotifier<int>
      crosshairRepaintNotifier; // Repaint notifier for crosshair container
  List<_MeasureWidgetContext>
      legendWidgetContext; // To measure legend size and position
  List<_ChartTemplateInfo> templates; // Chart Template info
  List<ChartAxis> zoomedAxisStates;
  Rect containerRect; // Contains chart container size
  bool zoomProgress;
  List<_ZoomAxisRange> zoomAxes;
  bool initialRender;
  List<_LegendRenderContext> legendToggleStates;
  List<ChartSegment> selectedSegments;
  List<ChartSegment> unselectedSegments;
  List<_MeasureWidgetContext> legendToggleTemplateStates;
  List<Rect> renderDatalabelRegions;
  Orientation deviceOrientation;
  List<Rect> dataLabelTemplateRegions;
  bool animateCompleted;

  void initializeDefaultValues() {
    controllerList = <AnimationController>[];
    zoomRepaintNotifier = ValueNotifier<int>(0);
    axisRepaintNotifier = ValueNotifier<int>(0);
    seriesRepaintNotifier = ValueNotifier<int>(0);
    tooltipRepaintNotifier = ValueNotifier<int>(0);
    trackballRepaintNotifier = ValueNotifier<int>(0);
    crosshairRepaintNotifier = ValueNotifier<int>(0);
    legendWidgetContext = <_MeasureWidgetContext>[];
    templates = <_ChartTemplateInfo>[];
    zoomedAxisStates = <ChartAxis>[];
    zoomAxes = <_ZoomAxisRange>[];
    containerRect = Rect.fromLTRB(0, 0, 0, 0);
    zoomProgress = false;
    initialRender = true;
    legendToggleStates = <_LegendRenderContext>[];
    selectedSegments = <ChartSegment>[];
    unselectedSegments = <ChartSegment>[];
    legendToggleTemplateStates = <_MeasureWidgetContext>[];
    renderDatalabelRegions = <Rect>[];
    dataLabelTemplateRegions = <Rect>[];
    animateCompleted = false;
  }

  @override
  void initState() {
    initializeDefaultValues();
    super.initState();
  }

  @override
  void didUpdateWidget(SfCartesianChart oldWidget) {
    _needsRepaintChart(widget, oldWidget);
    if (widget._chartState != null &&
        widget._chartState.legendWidgetContext != null &&
        widget._chartState.legendWidgetContext.isNotEmpty) {
      widget._chartState.legendWidgetContext.clear();
    }
    initialRender = false;
    super.didUpdateWidget(oldWidget);
  }

  void refresh() {
    if (widget._chartState.legendWidgetContext.isNotEmpty) {
      for (int i = 0; i < widget._chartState.legendWidgetContext.length; i++) {
        final _MeasureWidgetContext templateContext =
            widget._chartState.legendWidgetContext[i];
        final RenderBox renderBox = templateContext.context.findRenderObject();
        templateContext.size = renderBox.size;
      }
      setState(() {});
    }
  }

  void _redraw() {
    widget._chartState.initialRender = false;
    setState(() {});
  }

  @override
  void dispose() {
    if (controllerList != null) {
      for (int index = 0; index < controllerList.length; index++) {
        controllerList[index]
            .dispose(); // Need to dispose all the animation controller instance otherwise it will cause memory leak
      }
    }
    super.dispose();
  }

  // void loaded(){
  //   print('ChartLoaded');
  // }

  @override
  Widget build(BuildContext context) {
    widget._chartState = this;
    final ThemeData theme = Theme.of(context);
    widget._chartTheme.initializeChartTheme(theme);
    findVisibleSeries();
    return _ChartContainer(
        child: Container(
      decoration: BoxDecoration(
          color: widget.backgroundColor,
          image: widget.backgroundImage != null
              ? DecorationImage(image: widget.backgroundImage, fit: BoxFit.fill)
              : null,
          border:
              Border.all(color: widget.borderColor, width: widget.borderWidth)),
      child: Container(
          margin: EdgeInsets.fromLTRB(widget.margin.left, widget.margin.top,
              widget.margin.right, widget.margin.bottom),
          child: Column(
            children: <Widget>[renderChartTitle(), _renderChartElements()],
          )),
    ));
  }

  ///
  /// To render the chart title
  Widget renderChartTitle() {
    Widget titleWidget;
    if (widget.title.text != null && widget.title.text.isNotEmpty) {
      final Paint titleBackground = Paint()
        ..color = widget.title.borderColor
        ..style = PaintingStyle.stroke
        ..strokeWidth = widget.title.borderWidth;
      final ChartTextStyle titleStyle = _getTextStyle(
          textStyle: widget.title.textStyle,
          background: titleBackground,
          fontColor: widget.title.textStyle.color ??
              widget._chartTheme.chartTitleColor);
      final TextStyle textStyle = TextStyle(
          color: titleStyle.color,
          fontSize: titleStyle.fontSize,
          fontFamily: titleStyle.fontFamily,
          fontStyle: titleStyle.fontStyle,
          fontWeight: titleStyle.fontWeight);
      titleWidget = Container(
        child: Container(
          child: Text(widget.title.text,
              overflow: TextOverflow.clip,
              textAlign: TextAlign.center,
              textScaleFactor: 1.2,
              style: textStyle),
          decoration: BoxDecoration(
              color: widget.title.backgroundColor ??
                  widget._chartTheme.titleBackgroundColor,
              border: Border.all(
                  color: widget.title.borderWidth == 0
                      ? Colors.transparent
                      : widget.title.borderColor ??
                          widget._chartTheme.chartTitleColor,
                  width: widget.title.borderWidth)),
        ),
        alignment: (widget.title.alignment == ChartAlignment.near)
            ? Alignment.topLeft
            : (widget.title.alignment == ChartAlignment.far)
                ? Alignment.topRight
                : (widget.title.alignment == ChartAlignment.center)
                    ? Alignment.topCenter
                    : Alignment.topCenter,
      );
    } else {
      titleWidget = Container();
    }
    return titleWidget;
  }

  ///
  /// To arrange the chart area and legend area based on the legend position
  Widget _renderChartElements() {
    return Expanded(
      child: LayoutBuilder(
          builder: (BuildContext context, BoxConstraints constraints) {
        Widget element;
        if (widget._chartState.deviceOrientation != null &&
            widget._chartState.deviceOrientation !=
                MediaQuery.of(context).orientation) {
          widget._chartState.deviceOrientation =
              MediaQuery.of(context).orientation;
        } else {
          widget._chartState.selectedSegments = <ChartSegment>[];
          widget._chartState.unselectedSegments = <ChartSegment>[];
          widget._chartState.deviceOrientation =
              MediaQuery.of(context).orientation;
        }
        final List<Widget> legendTemplates = bindLegendTemplateWidgets();
        if (legendTemplates.isNotEmpty && legendWidgetContext.isEmpty) {
          element = Container(child: Stack(children: legendTemplates));
          SchedulerBinding.instance.addPostFrameCallback((_) => refresh());
        } else {
          initialize(constraints);
          widget._chartLegend
              ._calculateLegendBounds(widget._chartLegend.chartSize);
          element =
              _getElements(widget, _ContainerArea(chart: widget), constraints);
        }
        return element;
      }),
    );
  }

  List<Widget> bindLegendTemplateWidgets() {
    Widget legendWidget;
    final List<Widget> templates = <Widget>[];
    if (widget.legend.isVisible && widget.legend.legendItemBuilder != null) {
      for (int i = 0; i < widget.series.length; i++) {
        final CartesianSeries<dynamic, dynamic> series = widget.series[i];
        if (series.isVisibleInLegend) {
          legendWidget = widget.legend
              .legendItemBuilder(series._seriesName, series, null, i);
          templates.add(_MeasureWidgetSize(
              chart: widget,
              seriesIndex: i,
              pointIndex: null,
              type: 'Legend',
              currentKey: GlobalKey(),
              currentWidget: legendWidget,
              opacityValue: 0.0));
        }
      }
    }
    return templates;
  }

  void initialize(BoxConstraints constraints) {
    widget._chartLegend.chart = widget;
    widget._chartSeries.chart = widget;
    widget._chartAxis._chartWidget = widget;
    final double width = constraints.maxWidth;
    final double height = constraints.maxHeight;
    final LegendPosition position = widget.legend.position;
    final double widthPadding =
        position == LegendPosition.left || position == LegendPosition.right
            ? 5
            : 0;
    final double heightPadding =
        position == LegendPosition.top || position == LegendPosition.bottom
            ? 5
            : 0;
    widget._chartLegend.chartSize =
        Size(width - widthPadding, height - heightPadding);
  }

  void findVisibleSeries() {
    widget._chartSeries.visibleSeries = <CartesianSeries<dynamic, dynamic>>[];
    for (int i = 0; i < widget.series.length; i++) {
      widget.series[i]._seriesName = widget.series[i].name ?? 'Series $i';
      if (widget._chartState.initialRender) {
        widget.series[i]._visible = widget.series[i].isVisible;
      } else {
        widget.series[i]._visible =
            checkWithLegendToggleState(i, widget.series[i]._visible);
      }
      widget._chartSeries.visibleSeries.add(widget.series[i]);
    }
  }

  bool checkWithLegendToggleState(int seriesIndex, bool seriesVisible) {
     bool seriesRender;
    if (widget.legend.legendItemBuilder != null) {
      final List<_MeasureWidgetContext> legendToggles =
          widget._chartState.legendToggleTemplateStates;
      if (legendToggles.isNotEmpty) {
        for (int j = 0; j < legendToggles.length; j++) {
          final _MeasureWidgetContext item = legendToggles[j];
          if (seriesIndex == item.seriesIndex) {
            seriesRender = false;
            break;
          }
        }
      }
    } else {
      if (widget._chartState.legendToggleStates.isNotEmpty) {
        for (int j = 0; j < widget._chartState.legendToggleStates.length; j++) {
          final _LegendRenderContext legendRenderContext =
              widget._chartState.legendToggleStates[j];
          if (seriesIndex == legendRenderContext.seriesIndex) {
            seriesRender = false;
            break;
          }
        }
      }
    }
    return seriesRender ?? true;
  }
}

// ignore: must_be_immutable
class _ContainerArea extends StatelessWidget {
  // ignore: prefer_const_constructors_in_immutables
  _ContainerArea({this.chart});
  SfCartesianChart chart;
  List<Widget> _chartWidgets;
  CartesianSeries<dynamic, dynamic> _series;
  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
        builder: (BuildContext context, BoxConstraints constraints) {
      return Container(
          decoration: const BoxDecoration(color: Colors.transparent),
          child: _initializeChart(constraints, context));
    });
  }

  Offset _zoomStartPosition;
  Widget _initializeChart(BoxConstraints constraints, BuildContext context) {
    _calculateSize(constraints);
    _calculateBounds();
    return Container(
        decoration: const BoxDecoration(color: Colors.transparent),
        child: _renderWidgets(constraints, context));
  }

  void _calculateSize(BoxConstraints constraints) {
    final double width = constraints.maxWidth;
    final double height = constraints.maxHeight;
    chart._chartState.containerRect = Rect.fromLTWH(0, 0, width, height);
  }

  void _calculateBounds() {
    chart._chartSeries.chart = chart;
    chart._chartAxis._chartWidget = chart;
    chart._chartSeries?.processData();
    chart._chartAxis?._measureAxesBounds();
  }

  void _createSegment(SfCartesianChart chart, int index) {
    final CartesianSeries<dynamic, dynamic> chartSeries =
        chart._chartSeries.visibleSeries[index];
    chartSeries._internalCreateSegments();
  }

  Widget _renderWidgets(BoxConstraints constraints, BuildContext context) {
    _chartWidgets = <Widget>[];
    chart._chartState.renderDatalabelRegions = <Rect>[];
    chart.zoomPanBehavior._setChart(chart);
    _bindAxisWidgets('outside');
    _bindSeriesWidgets();
    _bindAxisWidgets('inside');
    _renderTemplates();
    _bindInteractionWidgets(constraints, context);
    return Container(child: Stack(children: _chartWidgets));
  }

  void _renderTemplates() {
    Widget labelWidget;
    _CartesianChartPoint<dynamic> point;
    chart._chartState.dataLabelTemplateRegions = <Rect>[];
    for (int i = 0; i < chart.series.length; i++) {
      final CartesianSeries<dynamic, dynamic> series = chart.series[i];
      if (series.dataLabelSettings.isVisible && series._visible) {
        for (int j = 0; j < series._dataPoints.length; j++) {
          point = series._dataPoints[j];
          if (point.isVisible && !point.isGap) {
            _calculateDataLabelPosition(series, point, j, chart);
            labelWidget = (series.dataLabelSettings.builder != null)
                ? series.dataLabelSettings
                    .builder(series.dataSource[j], point, series, i, j)
                : null;
            final _ChartTemplateInfo templateInfo = _ChartTemplateInfo(
                key: GlobalKey(),
                templateType: 'DataLabel',
                pointIndex: j,
                seriesIndex: i,
                clipRect: chart._chartAxis._axisClipRect,
                animationDuration: (series.animationDuration + 1000.0).floor(),
                widget: labelWidget,
                location: Offset(point.markerPoint.x, point.markerPoint.y));
            _replaceDataLabelTemplates(templateInfo);
          }
        }
      }
    }
    if (chart.annotations != null && chart.annotations.isNotEmpty) {
      for (int i = 0; i < chart.annotations.length; i++) {
        final CartesianChartAnnotation annotation = chart.annotations[i];
        final _ChartLocation location =
            _getAnnotationLocation(annotation, chart);
        final _ChartTemplateInfo chartTemplateInfo = _ChartTemplateInfo(
            key: GlobalKey(),
            animationDuration: 200,
            widget: annotation.child,
            templateType: 'Annotation',
            needMeasure: true,
            verticalAlignment: annotation.verticalAlignment,
            horizontalAlignment: annotation.horizontalAlignment,
            clipRect: annotation.region == AnnotationRegion.chart
                ? chart._chartState.containerRect
                : chart._chartAxis._axisClipRect,
            location: Offset(location.x.toDouble(), location.y.toDouble()));
        _replaceAnnotationTemplates(chartTemplateInfo);
      }
    }

    if (chart._chartState.templates.isNotEmpty) {
      final int templateLength = chart._chartState.templates.length;
      for (int i = 0; i < chart._chartState.templates.length; i++) {
        final _ChartTemplateInfo templateInfo = chart._chartState.templates[i];
        _chartWidgets.add(_RenderTemplate(
            template: templateInfo,
            templateIndex: i,
            templateLength: templateLength,
            chart: chart));
      }
    }
  }

  void _replaceAnnotationTemplates(_ChartTemplateInfo currentTemplate) {
    bool isReplaced = false;
    if (chart._chartState.templates.isNotEmpty) {
      for (int i = 0; i < chart._chartState.templates.length; i++) {
        final _ChartTemplateInfo templateInfo = chart._chartState.templates[i];
        if (currentTemplate.widget != null &&
            templateInfo.widget != null &&
            currentTemplate.widget.runtimeType.hashCode ==
                templateInfo.widget.runtimeType.hashCode) {
          templateInfo.needMeasure = true;
          templateInfo.widget = currentTemplate.widget;
          templateInfo.animationDuration = 0;
          templateInfo.clipRect = currentTemplate.clipRect;
          templateInfo.location = currentTemplate.location;
          templateInfo.horizontalAlignment =
              currentTemplate.horizontalAlignment;
          templateInfo.verticalAlignment = currentTemplate.verticalAlignment;
          isReplaced = true;
          break;
        } else if (currentTemplate.widget == null) {
          chart._chartState.templates.removeAt(i);
        }
      }
    }
    if (!isReplaced && currentTemplate.widget != null) {
      currentTemplate.needMeasure = true;
      chart._chartState.templates.add(currentTemplate);
    }
  }

  void _replaceDataLabelTemplates(_ChartTemplateInfo currentTemplate) {
    bool isReplaced = false;
    if (chart._chartState.templates.isNotEmpty) {
      for (int i = 0; i < chart._chartState.templates.length; i++) {
        final _ChartTemplateInfo templateInfo = chart._chartState.templates[i];
        if (templateInfo.pointIndex == currentTemplate.pointIndex &&
            templateInfo.seriesIndex == currentTemplate.seriesIndex) {
          if (currentTemplate.widget != null &&
              templateInfo.widget != null &&
              currentTemplate.widget.runtimeType.hashCode ==
                  templateInfo.widget.runtimeType.hashCode) {
            templateInfo.needMeasure = false;
            templateInfo.animationDuration = 0;
            templateInfo.clipRect = currentTemplate.clipRect;
            templateInfo.location = currentTemplate.location;
            isReplaced = true;
            break;
          } else if (currentTemplate.widget == null) {
            chart._chartState.templates.removeAt(i);
          }
        }
      }
    }
    if (!isReplaced && currentTemplate.widget != null) {
      currentTemplate.needMeasure = true;
      chart._chartState.templates.add(currentTemplate);
    }
  }

  void _bindSeriesWidgets() {
    Animation<double> chartElementAnimation;
    Animation<double> seriesAnimation;
    for (int i = 0; i < chart._chartSeries.visibleSeries.length; i++) {
      _series = chart._chartSeries.visibleSeries[i];
      if (_series != null) {
        _series._renderSeries(chart, i);
        _createSegment(chart, i);
        if (_series.selectionSettings != null) {
          _series.selectionSettings._selectionRenderer ??= _SelectionRenderer();
          _series.selectionSettings._selectionRenderer.chart = chart;
          _series.selectionSettings._selectionRenderer.selectedSegments =
              chart._chartState.selectedSegments;
          _series.selectionSettings._selectionRenderer.unselectedSegments =
              chart._chartState.unselectedSegments;
          if (chart.initialSelectedDataIndexes.isNotEmpty) {
            _series.selectionSettings
                ._selectedDataPointIndex(chart.initialSelectedDataIndexes);
          }
        }
        if (_series.animationDuration > 0 &&
            chart._chartState != null &&
            chart._chartState.initialRender) {
          chart._chartState.animationController =
              AnimationController(vsync: chart._chartState)
                ..addListener(() {
                  chart._chartState.seriesRepaintNotifier.value++;
                });

          chart._chartState.animationController.duration =
              Duration(milliseconds: _series.animationDuration.toInt());
          seriesAnimation =
              Tween<double>(begin: 0.0, end: 1.0).animate(CurvedAnimation(
            parent: chart._chartState.animationController,
            curve: const Interval(0.1, 0.8, curve: Curves.decelerate),
          ));

          chartElementAnimation =
              Tween<double>(begin: 0.0, end: 1.0).animate(CurvedAnimation(
            parent: chart._chartState.animationController,
            curve: const Interval(0.85, 1.0, curve: Curves.decelerate),
          ));
          chart._chartState.animationController.forward(from: 0.0);
          chart._chartState.controllerList.add(chart._chartState.animationController);
        }
        else {
          chart._chartState.animateCompleted = true;
        }
      }
      _chartWidgets.add(Container(
          child: RepaintBoundary(
              child: CustomPaint(
        painter: _getSeriesPainter(i, chart._chartState.animationController,
            seriesAnimation, chartElementAnimation),
      ))));

      _chartWidgets.add(Container(
          child: RepaintBoundary(
              child: CustomPaint(
                  painter: _ZoomRectPainter(
                      isRepaint: true,
                      chart: chart,
                      notifier: chart._chartState.zoomRepaintNotifier)))));
    }
    _animateComplete(chart._chartState.controllerList, chart);
  }

  void _bindAxisWidgets(String renderType) {
    if (chart._chartAxis._axisCollections != null &&
        chart._chartAxis._axisCollections.isNotEmpty &&
        chart._chartAxis._axisCollections.length > 1) {
      _chartWidgets.add(Container(
        child: RepaintBoundary(
            child: CustomPaint(
                painter: _CartesianAxesPainter(
                    chart: chart,
                    renderType: renderType,
                    isRepaint: chart._chartAxis._needsRepaint,
                    notifier: chart._chartState.axisRepaintNotifier))),
      ));
    }
  }

  void _bindInteractionWidgets(
      BoxConstraints constraints, BuildContext context) {
    final RenderBox renderBox = context.findRenderObject();
    _TrackballPainter trackballPainter;
    _CrosshairPainter crosshairPainter;
    Offset tapDownDetails;
    Offset touchPosition;
    final _TooltipPainter tooltipPainter = _TooltipPainter(
        valueNotifier: chart._chartState.tooltipRepaintNotifier);
    tooltipPainter._chartState = chart._chartState;
    tooltipPainter._chart = chart;
    tooltipPainter.tooltip = chart.tooltipBehavior;
    chart.tooltipBehavior._painter = tooltipPainter;
    final _ZoomRectPainter zoomRectPainter = _ZoomRectPainter(chart: chart);
    chart.zoomPanBehavior._painter = zoomRectPainter;

    CartesianSeries<dynamic, dynamic> selectionSeries;
    CartesianSeries<dynamic, dynamic> series;
    CartesianSeries<dynamic, dynamic> _findSeries(Offset position) {
      outerLoop:
      for (int i = chart._chartSeries.visibleSeries.length - 1; i >= 0; i--) {
        series = chart._chartSeries.visibleSeries[i];
        if (series._seriesType == 'column' ||
            series._seriesType == 'bar' ||
            series._seriesType == 'scatter' ||
            series._seriesType == 'bubble' ||
            series._seriesType == 'fastline' ||
            series._seriesType == 'area') {
          for (int j = 0; j < series._dataPoints.length; j++) {
            if (series._dataPoints[j].region.contains(position)) {
              break outerLoop;
            }
          }
        } else {
          bool isSelect = false;
          series = chart._chartSeries.visibleSeries[i];
          for (int k = chart._chartSeries.visibleSeries.length - 1;
              k >= 0;
              k--) {
            isSelect = series.selectionSettings._selectionRenderer
                ._isSeriesContainsPoint(
                    chart._chartSeries.visibleSeries[i], position);
            if (isSelect) {
              return chart._chartSeries.visibleSeries[i];
            }
          }
        }
      }
      return series;
    }

    if (chart.tooltipBehavior.enable) {
      if (chart.tooltipBehavior.builder != null) {
        chart.tooltipBehavior._tooltipTemplate = _TooltipTemplate(
            show: false, clipRect: chart._chartAxis._axisClipRect);
        _chartWidgets.add(chart.tooltipBehavior._tooltipTemplate);
      } else {
        _chartWidgets.add(Container(
            height: constraints.maxHeight,
            width: constraints.maxWidth,
            decoration: const BoxDecoration(color: Colors.transparent),
            child: CustomPaint(painter: tooltipPainter)));
      }
    }
    if (chart.trackballBehavior != null &&
        chart.trackballBehavior.enable) {
      trackballPainter = _TrackballPainter(
          chart: chart,
          valueNotifier: chart._chartState.trackballRepaintNotifier);
      chart.trackballBehavior._trackballPainter = trackballPainter;
      _chartWidgets.add(Container(
          height: constraints.maxHeight,
          width: constraints.maxWidth,
          decoration: const BoxDecoration(color: Colors.transparent),
          child: CustomPaint(painter: trackballPainter)));
    }
    if (chart.crosshairBehavior != null &&
        chart.crosshairBehavior.enable) {
      crosshairPainter = _CrosshairPainter(
          chart: chart,
          valueNotifier: chart._chartState.crosshairRepaintNotifier);
      chart.crosshairBehavior._crosshairPainter = crosshairPainter;
      _chartWidgets.add(Container(
          height: constraints.maxHeight,
          width: constraints.maxWidth,
          decoration: const BoxDecoration(color: Colors.transparent),
          child: CustomPaint(painter: crosshairPainter)));
    }
    _chartWidgets.add(Listener(
        onPointerDown: (PointerDownEvent event) {
          tapDownDetails = event.position;
          if (chart.zoomPanBehavior.enablePinching == true) {
            if (chart.zoomPanBehavior._touchStartPositions.length < 2) {
              chart.zoomPanBehavior._touchStartPositions.add(event);
            }
            if (chart.zoomPanBehavior._touchStartPositions.length == 2) {
              for (int axisIndex = 0;
                  axisIndex < chart._chartAxis._axisCollections.length;
                  axisIndex++) {
                final dynamic axis =
                    chart._chartAxis._axisCollections[axisIndex];
                chart.zoomPanBehavior.onPinchStart(
                    axis,
                    chart.zoomPanBehavior._touchStartPositions[0].position
                        .dx,
                    chart.zoomPanBehavior._touchStartPositions[0].position
                        .dy,
                    chart.zoomPanBehavior._touchStartPositions[1].position
                        .dx,
                    chart.zoomPanBehavior._touchStartPositions[1].position
                        .dy,
                    axis._zoomFactor);
              }
            }
          }

          final Offset position = renderBox.globalToLocal(event.position);
          touchPosition = position;
          if (chart._chartSeries.visibleSeries!= null && chart._chartSeries.visibleSeries.isNotEmpty && chart.selectionGesture == ActivationMode.singleTap && chart.zoomPanBehavior._isPinching != true) {
            selectionSeries = _findSeries(position);
            selectionSeries.selectionSettings._selectionRenderer.series =
                selectionSeries;
            selectionSeries.selectionSettings
                .onTouchDown(position.dx, position.dy);
          }
          if (chart.trackballBehavior != null &&
              chart.trackballBehavior.enable &&
              chart.trackballBehavior.activationMode ==
                  ActivationMode.singleTap) {
            chart.trackballBehavior.onTouchDown(position.dx, position.dy);
          }
          if (chart.crosshairBehavior != null &&
              chart.crosshairBehavior.enable &&
              chart.crosshairBehavior.activationMode ==
                  ActivationMode.singleTap) {
            chart.crosshairBehavior.onTouchDown(position.dx, position.dy);
          }
        },
        onPointerMove: (PointerMoveEvent event) {
          if (chart.zoomPanBehavior.enablePinching == true &&
              chart.zoomPanBehavior._touchStartPositions.length == 2) {
            chart.zoomPanBehavior._isPinching = true;
            final int pointerID = event.pointer;
            bool addPointer = true;
            for (int i = 0;
                i < chart.zoomPanBehavior._touchMovePositions.length;
                i++) {
              if (chart.zoomPanBehavior._touchMovePositions[i].pointer ==
                  pointerID) {
                addPointer = false;
              }
            }
            if (chart.zoomPanBehavior._touchMovePositions.length < 2 &&
                addPointer) {
              chart.zoomPanBehavior._touchMovePositions.add(event);
            }

            if (chart.zoomPanBehavior._touchMovePositions.length == 2 &&
                chart.zoomPanBehavior._touchStartPositions.length == 2) {
              if (chart.zoomPanBehavior._touchMovePositions[0].pointer ==
                  event.pointer) {
                chart.zoomPanBehavior._touchMovePositions[0] = event;
              }
              if (chart.zoomPanBehavior._touchMovePositions[1].pointer ==
                  event.pointer) {
                chart.zoomPanBehavior._touchMovePositions[1] = event;
              }
              chart.zoomPanBehavior._performPinchZooming(
                  chart.zoomPanBehavior._touchStartPositions,
                  chart.zoomPanBehavior._touchMovePositions);
            }
          }
        },
        onPointerUp: (PointerUpEvent event) {
          if (chart.zoomPanBehavior._touchStartPositions.length == 2 &&
              chart.zoomPanBehavior._touchMovePositions.length == 2 &&
              chart.zoomPanBehavior._isPinching) {
            for (int axisIndex = 0;
                axisIndex < chart._chartAxis._axisCollections.length;
                axisIndex++) {
              final ChartAxis axis =
                  chart._chartAxis._axisCollections[axisIndex];
              chart.zoomPanBehavior.onPinchEnd(
                  axis,
                  chart.zoomPanBehavior._touchMovePositions[0].position.dx,
                  chart.zoomPanBehavior._touchMovePositions[0].position.dy,
                  chart.zoomPanBehavior._touchMovePositions[1].position.dx,
                  chart.zoomPanBehavior._touchMovePositions[1].position.dy,
                  axis._zoomFactor);
            }
          }
          chart.zoomPanBehavior._touchStartPositions = <PointerEvent>[];
          chart.zoomPanBehavior._touchMovePositions = <PointerEvent>[];
          chart.zoomPanBehavior._isPinching = false;
          chart.zoomPanBehavior._delayRedraw = false;
          final Offset position = renderBox.globalToLocal(event.position);
          if (chart.tooltipBehavior.enable &&
              chart.tooltipBehavior.activationMode ==
                  ActivationMode.singleTap &&
              !chart.crosshairBehavior.enable) {
            if (chart.tooltipBehavior.builder != null) {
              const double padding = 15;
              if (chart._chartAxis._axisClipRect.contains(position)) {
                bool isTooltipRegion = false;
                for (int i = 0;
                    i < chart._chartSeries.visibleSeries.length;
                    i++) {
                  final CartesianSeries<dynamic, dynamic> series =
                      chart._chartSeries.visibleSeries[i];
                  int j = 0;
                  if (series.enableTooltip && series._regionalData != null) {
                    series._regionalData
                        .forEach((dynamic regionRect, dynamic values) {
                      final Rect region = regionRect[0];
                      final double left = region.left - padding;
                      final double right = region.right + padding;
                      final double top = region.top - padding;
                      final double bottom = region.bottom + padding;
                      final Rect paddedRegion =
                          Rect.fromLTRB(left, top, right, bottom);
                      if (paddedRegion.contains(position)) {
                        final Offset tooltipPosition = regionRect[1];
                        chart.tooltipBehavior._tooltipTemplate.rect =
                            Rect.fromLTWH(
                                tooltipPosition.dx,
                                tooltipPosition.dy - padding / 2,
                                region.width,
                                region.height);
                        chart.tooltipBehavior._tooltipTemplate.template =
                            chart.tooltipBehavior.builder(
                                series.dataSource[j],
                                series._dataPoints[j],
                                series,
                                j,
                                i);
                        isTooltipRegion = true;
                        chart.tooltipBehavior._tooltipTemplate.show =
                            isTooltipRegion;
                        chart.tooltipBehavior._tooltipTemplate?.state
                            ?._performTooltip();
                      }
                      j++;
                    });
                  }
                }
                if (!isTooltipRegion) {
                  chart.tooltipBehavior._tooltipTemplate.show =
                      isTooltipRegion;
                  chart.tooltipBehavior._tooltipTemplate?.state
                      ?._performTooltip();
                }
              }
            } else {
              chart.tooltipBehavior.onTouchUp(position.dx, position.dy);
            }
          }
          if ((chart.trackballBehavior != null &&
                  chart.trackballBehavior.enable && chart.trackballBehavior.shouldAlwaysShow !=null &&
              !chart.trackballBehavior.shouldAlwaysShow &&
                  chart.trackballBehavior.activationMode !=
                      ActivationMode.doubleTap && chart.zoomPanBehavior._isPinching != true) ||
              (chart.zoomPanBehavior != null &&
                  (chart.zoomPanBehavior.enableDoubleTapZooming ||
                  chart.zoomPanBehavior.enablePanning ||
                  chart.zoomPanBehavior.enablePinching ||
                  chart.zoomPanBehavior.enableSelectionZooming &&
                      chart.trackballBehavior.shouldAlwaysShow))) {
            chart.trackballBehavior.onTouchUp(position.dx, position.dy);
            chart.trackballBehavior._isLongPressActivated = false;
          }
          if ((chart.crosshairBehavior != null &&
                  chart.crosshairBehavior.enable &&
                  !chart.crosshairBehavior.shouldAlwaysShow &&
                  chart.crosshairBehavior.activationMode !=
                      ActivationMode.doubleTap && chart.zoomPanBehavior._isPinching != true ) ||
              (chart.zoomPanBehavior != null &&
                  (chart.zoomPanBehavior.enableDoubleTapZooming ||
                      chart.zoomPanBehavior.enablePanning ||
                      chart.zoomPanBehavior.enablePinching ||
                      chart.zoomPanBehavior.enableSelectionZooming &&
                          chart.crosshairBehavior.shouldAlwaysShow))) {
            chart.crosshairBehavior.onTouchUp(position.dx, position.dy);
            chart.crosshairBehavior._isLongPressActivated = false;
          }
        },
        child: GestureDetector(
            onTapDown: (TapDownDetails details) {
              final Offset position =
                  renderBox.globalToLocal(details.globalPosition);
              touchPosition = position;
              if (chart.onPointTapped != null) {
                _getPointSeriesIndex(chart, position);
              }
              if (chart.onAxisLabelTapped != null) {
                _triggerAxisLabelEvent(position);
              }
            },
            onDoubleTap: () {
              if (tapDownDetails != null) {
                final Offset position = renderBox.globalToLocal(tapDownDetails);
                if (chart.tooltipBehavior.enable &&
                    chart.tooltipBehavior.activationMode ==
                        ActivationMode.doubleTap &&
                    !chart.crosshairBehavior.enable) {
                  chart.tooltipBehavior
                      .onDoubleTap(position.dx, position.dy);
                }
                if (chart.trackballBehavior != null &&
                    chart.trackballBehavior.enable &&
                    chart.trackballBehavior.activationMode ==
                        ActivationMode.doubleTap) {
                  chart.trackballBehavior
                      .onDoubleTap(position.dx, position.dy);
                }
                if (chart.crosshairBehavior != null &&
                    chart.crosshairBehavior.enable &&
                    chart.crosshairBehavior.activationMode ==
                        ActivationMode.doubleTap) {
                  chart.crosshairBehavior
                      .onDoubleTap(position.dx, position.dy);
                }
                if (chart._chartSeries.visibleSeries!= null && chart.selectionGesture == ActivationMode.doubleTap) {
                  selectionSeries = _findSeries(position);
                  selectionSeries.selectionSettings._selectionRenderer.series =
                      selectionSeries;
                  selectionSeries.selectionSettings
                      .onDoubleTap(position.dx, position.dy);
                }
              }

              if (chart.zoomPanBehavior.enableDoubleTapZooming == true) {
                final Offset doubleTapPosition = touchPosition;
                final Offset position = doubleTapPosition;
                if (position != null)
                  chart.zoomPanBehavior.onDoubleTap(position.dx,
                      position.dy, chart.zoomPanBehavior._zoomFactor);
              }
            },
            onLongPressMoveUpdate: (LongPressMoveUpdateDetails details) {
              Offset position;
              if (chart.zoomPanBehavior._isPinching != true) {
                position = renderBox.globalToLocal(details.globalPosition);
                if (chart.zoomPanBehavior.enableSelectionZooming &&
                    position != null &&
                    _zoomStartPosition != null) {
                  chart.zoomPanBehavior._canPerformSelection = true;
                  chart.zoomPanBehavior.onDrawSelectionZoomRect(
                      position.dx,
                      position.dy,
                      _zoomStartPosition.dx,
                      _zoomStartPosition.dy);
                }
              }
              if (chart.trackballBehavior != null &&
                  chart.trackballBehavior.enable &&
                  chart.trackballBehavior.activationMode !=
                      ActivationMode.doubleTap) {
                if (chart.trackballBehavior.activationMode ==
                    ActivationMode.singleTap) {
                  chart.trackballBehavior
                      .onTouchMove(position.dx, position.dy);
                } else if (chart.trackballBehavior != null &&
                    chart.trackballBehavior.activationMode ==
                        ActivationMode.longPress &&
                    chart.trackballBehavior._isLongPressActivated) {
                  chart.trackballBehavior
                      .onTouchMove(position.dx, position.dy);
                }
              }
              if (chart.crosshairBehavior != null &&
                  chart.crosshairBehavior.enable &&
                  chart.crosshairBehavior.activationMode !=
                      ActivationMode.doubleTap) {
                if (chart.crosshairBehavior.activationMode ==
                    ActivationMode.singleTap) {
                  chart.crosshairBehavior
                      .onTouchMove(position.dx, position.dy);
                } else if (chart.crosshairBehavior != null &&
                    chart.crosshairBehavior.activationMode ==
                        ActivationMode.longPress &&
                    chart.crosshairBehavior._isLongPressActivated) {
                  chart.crosshairBehavior
                      .onTouchMove(position.dx, position.dy);
                }
              }
            },
            onLongPress: () {
              Offset position;
              if (tapDownDetails != null) {
                position = renderBox.globalToLocal(tapDownDetails);
                if (chart.tooltipBehavior.enable &&
                    chart.tooltipBehavior.activationMode ==
                        ActivationMode.longPress) {
                  chart.tooltipBehavior
                      .onLongPress(position.dx, position.dy);
                }
              }
              if (chart._chartSeries.visibleSeries!= null && chart.selectionGesture == ActivationMode.longPress) {
                selectionSeries = _findSeries(position);
                selectionSeries.selectionSettings._selectionRenderer.series =
                    selectionSeries;
                selectionSeries.selectionSettings
                    .onLongPress(position.dx, position.dy);
              }
              if (chart.trackballBehavior != null &&
                  chart.trackballBehavior.enable == true &&
                  chart.trackballBehavior.activationMode ==
                      ActivationMode.longPress) {
                chart.trackballBehavior._isLongPressActivated = true;
                chart.trackballBehavior
                    .onTouchDown(position.dx, position.dy);
              }
              if (chart.crosshairBehavior != null &&
                  chart.crosshairBehavior.activationMode ==
                      ActivationMode.longPress) {
                chart.crosshairBehavior._isLongPressActivated = true;
                chart.crosshairBehavior
                    .onTouchDown(position.dx, position.dy);
              }
            },
            onLongPressEnd: (LongPressEndDetails details) {
              if (chart.zoomPanBehavior._isPinching != true) {
                chart.zoomPanBehavior._canPerformSelection = false;
                if (chart.zoomPanBehavior.enableSelectionZooming) {
                  chart.zoomPanBehavior._doSelectionZooming(
                      chart.zoomPanBehavior._zoomingRect);
                  if (chart.zoomPanBehavior._canPerformSelection != true) {
                    chart.zoomPanBehavior._zoomingRect =
                        Rect.fromLTRB(0, 0, 0, 0);
                  }
                }
              }
            },
            onPanUpdate: (DragUpdateDetails details) {
              Offset position;
              if (chart.zoomPanBehavior._isPinching != true) {
                position = renderBox.globalToLocal(details.globalPosition);
                if (chart.zoomPanBehavior._isPanning == true &&
                    chart.zoomPanBehavior.enablePanning &&
                    chart.zoomPanBehavior._previousMovedPosition != null) {
                  chart.zoomPanBehavior.onPan(position.dx, position.dy);
                }
                chart.zoomPanBehavior._previousMovedPosition = position;
              }
              if (chart.trackballBehavior != null &&
                  chart.trackballBehavior.enable &&
                  chart.trackballBehavior.activationMode !=
                      ActivationMode.doubleTap) {
                if (chart.trackballBehavior.activationMode ==
                    ActivationMode.singleTap) {
                  chart.trackballBehavior
                      .onTouchMove(position?.dx, position?.dy);
                } else if (chart.trackballBehavior != null &&
                    chart.trackballBehavior.activationMode ==
                        ActivationMode.longPress &&
                    chart.trackballBehavior._isLongPressActivated ==
                        true) {
                  chart.trackballBehavior
                      .onTouchMove(position?.dx, position?.dy);
                }
              }
              if (chart.crosshairBehavior != null &&
                  chart.crosshairBehavior.enable &&
                  chart.crosshairBehavior.activationMode !=
                      ActivationMode.doubleTap) {
                if (chart.crosshairBehavior.activationMode ==
                    ActivationMode.singleTap) {
                  chart.crosshairBehavior
                      .onTouchMove(position?.dx, position?.dy);
                } else if (chart.crosshairBehavior != null &&
                    chart.crosshairBehavior.activationMode ==
                        ActivationMode.longPress &&
                    chart.crosshairBehavior._isLongPressActivated) {
                  chart.crosshairBehavior
                      .onTouchMove(position?.dx, position?.dy);
                }
              }
            },
            onPanEnd: (DragEndDetails details) {
              if (chart.zoomPanBehavior._isPinching != true) {
                chart.zoomPanBehavior._isPanning = false;
                chart.zoomPanBehavior._previousMovedPosition = null;
              }
              if (chart.trackballBehavior.enable &&
                  !chart.trackballBehavior.shouldAlwaysShow &&
                  chart.trackballBehavior.activationMode !=
                      ActivationMode.doubleTap) {
                chart.trackballBehavior
                    .onTouchUp(touchPosition?.dx, touchPosition?.dy);
                chart.trackballBehavior._isLongPressActivated = false;
              }
              if (chart.crosshairBehavior.enable &&
                  !chart.crosshairBehavior.shouldAlwaysShow &&
                  chart.crosshairBehavior.activationMode !=
                      ActivationMode.doubleTap) {
                chart.crosshairBehavior
                    .onTouchUp(touchPosition?.dx, touchPosition?.dy);
              }
            },
            onPanDown: (DragDownDetails details) {
              if (chart.zoomPanBehavior._isPinching != true) {
                _zoomStartPosition =
                    renderBox.globalToLocal(details.globalPosition);
                if (chart.zoomPanBehavior.enablePanning == true) {
                  chart.zoomPanBehavior._isPanning = true;
                  chart.zoomPanBehavior._previousMovedPosition = null;
                }
              }
            },
            child: Container(
                height: constraints.maxHeight,
                width: constraints.maxWidth,
                decoration: const BoxDecoration(color: Colors.transparent)))));
  }

  void _getPointSeriesIndex(SfCartesianChart chart, Offset position) {
    for (int i = 0; i < chart._chartSeries.visibleSeries.length; i++) {
      final CartesianSeries<dynamic, dynamic> series =
          chart._chartSeries.visibleSeries[i];
      for (int k = 0; k < series.segments.length; k++) {
        if (series.segments[k].segmentRect.contains(position)) {
          PointTapArgs pointTapArgs;
          pointTapArgs = PointTapArgs();
          pointTapArgs.pointIndex = k;
          pointTapArgs.seriesIndex = i;
          pointTapArgs.series = series;
          chart.onPointTapped(pointTapArgs);
        }
      }
    }
  }

  void _triggerAxisLabelEvent(Offset position) {
    for (int i = 0; i < chart._chartAxis._axisCollections.length; i++) {
      final List<AxisLabel> labels =
          chart._chartAxis._axisCollections[i]._visibleLabels;
      for (int k = 0; k < labels.length; k++) {
        if (chart._chartAxis._axisCollections[i].isVisible &&
            labels[k]._labelRegion.contains(position)) {
          AxisLabelTapArgs labelArgs;
          labelArgs = AxisLabelTapArgs();
          labelArgs.text = labels[k].text;
          labelArgs.axis = chart._chartAxis._axisCollections[i];
          labelArgs.value = labels[k].value;
          chart.onAxisLabelTapped(labelArgs);
        }
      }
    }
  }

  CustomPainter _getSeriesPainter(
      int value,
      AnimationController controller,
      Animation<double> seriesAnimation,
      Animation<double> chartElementAnimation) {
    CustomPainter customPainter;
    final CartesianSeries<dynamic, dynamic> series = chart.series[value];
    if (series._seriesType == 'line') {
      customPainter = _LineChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'spline') {
      customPainter = _SplineChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'column') {
      customPainter = _ColumnChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'scatter') {
      customPainter = _ScatterChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'stepline') {
      customPainter = _StepLineChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'area') {
      customPainter = _AreaChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'bubble') {
      customPainter = _BubbleChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'bar') {
      customPainter = _BarChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    } else if (series._seriesType == 'fastline') {
      customPainter = _FastLineChartPainter(
          chart: chart,
          series: series,
          isRepaint: series._needsRepaint,
          animationController: controller,
          seriesAnimation: seriesAnimation,
          chartElementAnimation: chartElementAnimation,
          notifier: chart._chartState.seriesRepaintNotifier);
    }
    return customPainter;
  }
}
