part of charts;

abstract class CartesianSeries<T, D> extends ChartSeries<T, D> {
  CartesianSeries(
      {this.xValueMapper,
      this.yValueMapper,
      this.dataLabelMapper,
      this.name,
      this.dataSource,
      this.xAxisName,
      this.yAxisName,
      this.sizeValueMapper,
      this.pointColorMapper,
      this.color,
      this.legendItemText,
      this.sortFieldValueMapper,
      this.gradient,
      this.width,
      MarkerSettings markerSettings,
      bool isVisible,
      bool enableTooltip,
      EmptyPointSettings emptyPointSettings,
      DataLabelSettings dataLabelSettings,
      double animationDuration,
      List<double> dashArray,
      Color borderColor,
      double borderWidth,
      SelectionSettings selectionSettings,
      bool isVisibleInLegend,
      LegendIconType legendIconType,
      double opacity,
      List<ChartSegment> segments,
      SortingOrder sortingOrder})
      : isVisible = isVisible ?? true,
        markerSettings = markerSettings ?? MarkerSettings(),
        dataLabelSettings = dataLabelSettings ?? DataLabelSettings(),
        enableTooltip = enableTooltip ?? true,
        emptyPointSettings = emptyPointSettings ?? EmptyPointSettings(),
        dashArray = dashArray ?? <double>[0, 0],
        animationDuration = animationDuration ?? 1500,
        borderColor = borderColor ?? Colors.transparent,
        selectionSettings = selectionSettings ?? SelectionSettings(),
        legendIconType = legendIconType ?? LegendIconType._seriesType,
        isVisibleInLegend = isVisibleInLegend ?? true,
        borderWidth = borderWidth ?? 0,
        opacity = opacity ?? 1,
        sortingOrder = sortingOrder ?? SortingOrder.none,
        segments = segments ?? <ChartSegment>[],
        super(
            name: name,
            xValueMapper: xValueMapper,
            yValueMapper: yValueMapper,
            sortFieldValueMapper: sortFieldValueMapper,
            pointColorMapper: pointColorMapper,
            dataLabelMapper: dataLabelMapper,
            dataSource: dataSource,
            emptyPointSettings: emptyPointSettings,
            dataLabelSettings: dataLabelSettings,
            enableTooltip: enableTooltip,
            animationDuration: animationDuration,
            selectionSettings: selectionSettings,
            legendIconType: legendIconType,
            sortingOrder: sortingOrder,
            opacity: opacity);

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  dataSource: chartData,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///final dynamic chartData = <SalesData>[
  ///    SalesData(DateTime(2005, 0, 1), 'India', 0.5, 16, 10, 680, 760),
  ///    SalesData(DateTime(2006, 0, 1), 'China', 1, 12, 7, 550, 880),
  ///    SalesData(DateTime(2007, 0, 1), 'USA', 1.5, 18, 11, 440, 788),
  ///    SalesData(DateTime(2008, 0, 1), 'Japan', 2, 14, 7, 350, 560),
  ///    SalesData(DateTime(2009, 0, 1), 'Russia', 2.5, 11, 8, 444, 566),
  ///    SalesData(DateTime(2010, 0, 1), 'France', 3, 14, 6, 780, 650),
  ///    SalesData(DateTime(2011, 0, 1), 'Germany', 3.5, 10, 3, 450, 800),
  ///    SalesData(DateTime(2012, 0, 1), 'Italy', 4, 16, 7, 450, 800)
  ///  ];
  ///```
  final List<T> dataSource;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  dataSource: chartData,
  ///                  xValueMapper: (SalesData sales, _) => sales.numeric,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///class SalesData {
  ///   SalesData(this.year, this.category, this.numeric, this.sales1, this.sales2, this.sales3, this.sales4);
  ///   final DateTime year;
  ///   final String category;
  ///   final double numeric;
  ///   final int sales1;
  ///   final int sales2;
  ///   final int sales3;
  ///   final int sales4;
  ///}
  ///```
  final ChartIndexedValueMapper<D> xValueMapper;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  dataSource: chartData,
  ///                  yValueMapper: (SalesData sales, _) => sales.sales3,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///class SalesData {
  ///   SalesData(this.year, this.category, this.numeric, this.sales1, this.sales2, this.sales3, this.sales4);
  ///   final DateTime year;
  ///   final String category;
  ///   final double numeric;
  ///   final int sales1;
  ///   final int sales2;
  ///   final int sales3;
  ///   final int sales4;
  ///}
  ///```
  final ChartIndexedValueMapper<num> yValueMapper;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataSource: chartData,
  ///                       pointColorMapper: (BubbleColors sales, _) => sales.pointColorMapper,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///final dynamic chartData = <BubbleColors>[
  ///    BubbleColors(92.2, 7.8, 1.347, const Color.fromRGBO(0, 0, 255, 1)),
  ///    BubbleColors(74, 6.5, 1.241, const Color.fromRGBO(255, 0, 0, 1)),
  ///    BubbleColors(90.4, 6.0, 0.238, const Color.fromRGBO(255, 100, 102, 1)),
  ///    BubbleColors(99.4, 2.2, 0.197, const Color.fromRGBO(122, 100, 255, 1)),
  /// ];
  ///class BubbleColors {
  ///  BubbleColors(this.year, this.growth,[this.bubbleSize, this.pointColorMapper]);
  ///  final num year;
  ///  final num growth;
  ///  final num bubbleSize;
  ///  final Color pointColorMapper;
  ///}
  ///```
  final ChartIndexedValueMapper<Color> pointColorMapper;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                 dataSource: <SalesData>[
  ///                    SalesData(DateTime(2005, 0, 1), 'India', Colors.red, 16, 10, 680, 760),
  ///                    SalesData(DateTime(2006, 0, 1), 'China', Colors.blue, 12, 7, 550, 880),
  ///                    SalesData(DateTime(2007, 0, 1), 'USA', Colors.green, 18, 11, 440, 788),
  ///                    SalesData(DateTime(2008, 0, 1), 'Japan', Colors.yellow, 14, 7, 350, 560),
  ///                ],
  ///                dataLabelMapper: (ChartData data, _) => data.category,
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///class SalesData {
  ///   SalesData(this.year, this.category, this.pointColor, this.sales1, this.sales2, this.sales3, this.sales4);
  ///   final DateTime year;
  ///   final String category;
  ///   final Color pointColor;
  ///   final int sales1;
  ///   final int sales2;
  ///   final int sales3;
  ///   final int sales4;
  ///}
  ///```
  final ChartIndexedValueMapper<String> dataLabelMapper;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataSource: chartData,
  ///                       sizeValueMapper: (BubbleColors sales, _) => sales.bubbleSize,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///final dynamic chartData = <BubbleColors>[
  ///    BubbleColors(92.2, 7.8, 1.347, const Color.fromRGBO(0, 0, 255, 1)),
  ///    BubbleColors(74, 6.5, 1.241, const Color.fromRGBO(255, 0, 0, 1)),
  ///    BubbleColors(90.4, 6.0, 0.238, const Color.fromRGBO(255, 100, 102, 1)),
  ///    BubbleColors(99.4, 2.2, 0.197, const Color.fromRGBO(122, 100, 255, 1)),
  /// ];
  ///class BubbleColors {
  ///  BubbleColors(this.year, this.growth,[this.bubbleSize, this.sizeValueMapper]);
  ///  final num year;
  ///  final num growth;
  ///  final num bubbleSize;
  ///  final Color pointColorMapper;
  ///}
  ///```
  final ChartIndexedValueMapper<num> sizeValueMapper;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///             axes: <ChartAxis>[
  ///                        NumericAxis(
  ///                            plotOffset: 0,
  ///                            majorGridLines: MajorGridLines(color: Colors.transparent),
  ///                            opposedPosition: true,
  ///                            name: 'xAxis1',
  ///                            axisTitle: AxisTitle(
  ///                                text: 'China - Growth'))
  ///                      ],
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataSource: chartData,
  ///                       xAxisName: 'xAxis1',
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///final dynamic chartData = <BubbleColors>[
  ///    BubbleColors(92.2, 7.8, 1.347, const Color.fromRGBO(0, 0, 255, 1)),
  ///    BubbleColors(74, 6.5, 1.241, const Color.fromRGBO(255, 0, 0, 1)),
  ///    BubbleColors(90.4, 6.0, 0.238, const Color.fromRGBO(255, 100, 102, 1)),
  ///    BubbleColors(99.4, 2.2, 0.197, const Color.fromRGBO(122, 100, 255, 1)),
  /// ];
  ///```
  final String xAxisName;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///             axes: <ChartAxis>[
  ///                        NumericAxis(
  ///                            plotOffset: 0,
  ///                            majorGridLines: MajorGridLines(color: Colors.transparent),
  ///                            opposedPosition: true,
  ///                            name: 'yAxis1',
  ///                            axisTitle: AxisTitle(
  ///                                text: 'China - Population '))
  ///                      ],
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataSource: chartData,
  ///                       yAxisName: 'yAxis1',
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///final dynamic chartData = <BubbleColors>[
  ///    BubbleColors(92.2, 7.8, 1.347, const Color.fromRGBO(0, 0, 255, 1)),
  ///    BubbleColors(74, 6.5, 1.241, const Color.fromRGBO(255, 0, 0, 1)),
  ///    BubbleColors(90.4, 6.0, 0.238, const Color.fromRGBO(255, 100, 102, 1)),
  ///    BubbleColors(99.4, 2.2, 0.197, const Color.fromRGBO(122, 100, 255, 1)),
  /// ];
  ///```
  final String yAxisName;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       color: const Color.fromRGBO(255, 0, 102, 1),
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final Color color;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       width: 2,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final double width;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       markerSettings: MarkerSettings(isVisible: true),
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final MarkerSettings markerSettings;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       emptyPointSettings: EmptyPointSettings(color: Colors.black),
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final EmptyPointSettings emptyPointSettings;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataLabelSettings: DataLabelSettings(isVisible: true),
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final DataLabelSettings dataLabelSettings;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, String>>[
  ///                   BarSeries<SalesData, String>(
  ///                       gradient: gradients,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///final List <Color> color = <Color>[];
  ///    color.add(Colors.pink[50]);
  ///    color.add(Colors.pink[200]);
  ///    color.add(Colors.pink);
  ///
  ///    final List<double> stops = <double>[];
  ///    stops.add(0.0);
  ///    stops.add(0.5);
  ///    stops.add(1.0);
  ///
  ///    final LinearGradient gradients = LinearGradient(colors: color, stops: stops);
  ///```
  final LinearGradient gradient;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       name: 'Bubble Series',
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final String name;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       enableTooltip: true,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final bool enableTooltip;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       dashArray: <double>[10, 10],
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final List<double> dashArray;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       animationDuration: 1500,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final double animationDuration;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       borderColor: Colors.red,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final Color borderColor;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       borderWidth: 5,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final double borderWidth;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       legendIconType: LegendIconType.diamond,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final LegendIconType legendIconType;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       isVisibleInLegend: true,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final bool isVisibleInLegend;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       legendItemText: 'legend',
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final String legendItemText;

  /// Gets the Segments collection variable declarations.
  final List<ChartSegment> segments;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final SelectionSettings selectionSettings;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                      opacity: 1
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final double opacity;

  /// Stores the series type
  String _seriesType;

  /// This property holds the sorting field in dataSource
  final ChartIndexedValueMapper<dynamic> sortFieldValueMapper;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                    sortingOrder: SortingOrder.ascending,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final SortingOrder sortingOrder;

  /// It specifies the show or hide the series.
  final bool isVisible;

  /// Holds the collection of cartesian data points
  List<_CartesianChartPoint<dynamic>> _dataPoints;

  /// Name of the series
  String _seriesName;

  /// Holds the information for x Axis
  ChartAxis _xAxis;

  /// Holds the information for y Axis
  ChartAxis _yAxis;

  /// Minimum x value for Series
  num _minimumX;

  /// Maximum x value for Series
  num _maximumX;

  /// Minimum y value for Series
  num _minimumY;

  /// Maximum y value for Series
  num _maximumY;

  /// Hold the data about point regions
  Map<dynamic, dynamic> _regionalData;

  /// Color for the series based on color palette
  Color _seriesColor;

  /// Hold the information about chart class
  SfCartesianChart _chart;

  /// Side by side index for series
  int _sideBySideIndex;

  /// Contains the collection of path for markers
  List<Path> _markerShapes;

  bool _visible;

  /// calculate the series region
  void _renderSeries(SfCartesianChart chart, int index) {
    _calculateSeriesRegion(chart, index);
  }

  /// Create segments for series
  void _internalCreateSegments() {
    if (segments.isNotEmpty) {
      segments?.clear();
    }
    createSegments();
  }

  /// Draw series
  void _draw(Canvas canvas, [double animationFactor]) {
    for (ChartSegment chartSegment in segments) {
      chartSegment.animationFactor = animationFactor;
      chartSegment.onPaint(canvas);
    }
  }

  /// Creates collection of segments for series
  void createSegments();

  /// To create segment for series
  ChartSegment createSegment();

  /// To customize each segments
  void customizeSegment(ChartSegment segment);

  /// To customize each data markers
  void drawDataMarker(int index, Canvas canvas, Paint fillPaint,
      Paint strokePaint, double pointX, double pointY);

  /// To customize each data labels
  void drawDataLabel(int index, Canvas canvas, String dataLabel, double pointX,
      double pointY, int angle, ChartTextStyle style);
}
