part of charts;

class DataLabelSettings {
  DataLabelSettings(
      {this.alignment = ChartAlignment.center,
      this.color,
      ChartTextStyle textStyle,
      EdgeInsets margin,
      this.opacity = 1,
      this.position = CartesianLabelPosition.auto,
      this.borderRadius = 5,
      this.isVisible = false,
      this.angle = 0,
      this.builder,
      this.useSeriesColor = false,
      Color borderColor,
      double borderWidth,
      LabelIntersectAction labelIntersectAction,
      ConnectorLineSettings connectorLineSettings,
      LabelPosition labelPosition})
      : borderColor = borderColor ?? Colors.transparent,
      labelPosition = labelPosition ?? LabelPosition.inside,
      borderWidth = borderWidth ?? 0,
      margin = margin ?? const EdgeInsets.fromLTRB(0,0,0,0),
      connectorLineSettings = connectorLineSettings ?? ConnectorLineSettings(),
      labelIntersectAction = labelIntersectAction ?? LabelIntersectAction.hide,
      textStyle = textStyle ?? ChartTextStyle();

  ///Works only for cartesian type series
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionGesture: ActivationMode.doubleTap,
  ///            series: <SplineSeries<SalesData, num>>[
  ///                SplineSeries<SalesData, num>(
  ///                  dataLabelSettings: DataLabelSettings(alignment: ChartAlignment.center),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final ChartAlignment alignment;
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionGesture: ActivationMode.doubleTap,
  ///            series: <SplineSeries<SalesData, num>>[
  ///                SplineSeries<SalesData, num>(
  ///                  dataLabelSettings: DataLabelSettings(color: Colors.red),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final Color color;
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionGesture: ActivationMode.doubleTap,
  ///            series: <SplineSeries<SalesData, num>>[
  ///                SplineSeries<SalesData, num>(
  ///                  dataLabelSettings: DataLabelSettings(textStyle: ChartTextStyle(color: Colors.red)),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final ChartTextStyle textStyle;
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionGesture: ActivationMode.doubleTap,
  ///            series: <SplineSeries<SalesData, num>>[
  ///                SplineSeries<SalesData, num>(
  ///                  dataLabelSettings: DataLabelSettings(
  ///                       margin: const EdgeInsets.all(2),
  ///                       borderColor: Colors.red,
  ///                       borderWidth: 2),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final EdgeInsets margin;
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            selectionGesture: ActivationMode.doubleTap,
  ///            series: <SplineSeries<SalesData, num>>[
  ///                SplineSeries<SalesData, num>(
  ///                  dataLabelSettings: DataLabelSettings(
  ///                       opacity: 0),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final double opacity;

  final bool useSeriesColor;

  ///Works only for cartesian type series
  final CartesianLabelPosition position;
  final double borderRadius;
  final bool isVisible;
  final int angle;
  final Color borderColor;
  final double borderWidth;

  ///Works only for circular type series
  final LabelPosition labelPosition;

  ///Works only for circular type series
  final ConnectorLineSettings connectorLineSettings;
  
  ///Works only for circular type series
  final LabelIntersectAction labelIntersectAction;
  
  final ChartWidgetBuilder<dynamic> builder;
}
