part of charts;

class CrosshairBehavior extends ChartBehavior {
  CrosshairBehavior({
    ActivationMode activationMode,
    CrosshairLineType lineType,
    this.lineDashArray,
    this.enable = false,
    this.lineColor,
    this.lineWidth = 1,
    this.shouldAlwaysShow = false,
  })  : activationMode = activationMode ?? ActivationMode.longPress,
        lineType = lineType ?? CrosshairLineType.both{
    _isLongPressActivated = false;
  }


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true),
  ///        ));
  ///}
  ///```
  final bool enable;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true, lineWidth: 5),
  ///        ));
  ///}
  ///```
  final double lineWidth;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true, lineColor: Colors.red),
  ///        ));
  ///}
  ///```
  final Color lineColor;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true, lineDashArray: [10,10]),
  ///        ));
  ///}
  ///```
  final List<double> lineDashArray;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true, activationMode: ActivationMode.doubleTap),
  ///        ));
  ///}
  ///```
  final ActivationMode activationMode;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true, lineType: CrosshairLineType.horizontal),
  ///        ));
  ///}
  ///```
  final CrosshairLineType lineType;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           crosshairBehavior: CrosshairBehavior(enable: true, shouldAlwaysShow: true),
  ///        ));
  ///}
  ///```
  final bool shouldAlwaysShow;

  /// Touch position
  Offset _position;

  /// Holds the instance of CrosshairPainter
  _CrosshairPainter _crosshairPainter;

  /// Check whether long press activated or not .
  bool _isLongPressActivated;

  // show method is used to show crosshair at the given position
  void show(double x, double y) {
    if (_crosshairPainter != null && activationMode!=ActivationMode.none && x !=null && y !=null) {
      _crosshairPainter.generateAllPoints(Offset(x, y));
      _crosshairPainter.canResetPath = false;
      _crosshairPainter.chart._chartState.crosshairRepaintNotifier.value++;
    }
  }

  // hide method is used to hide the crosshair
  void hide() {
    if (_crosshairPainter != null) {
      _crosshairPainter.canResetPath = true;
      _crosshairPainter.chart._chartState.crosshairRepaintNotifier.value++;
    }
  }

  @override
  void onDoubleTap(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onLongPress(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onTouchDown(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onTouchMove(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onTouchUp(double xPos, double yPos) {
    hide();
  }

  @override
  void onPaint(Canvas canvas) {
    if (_crosshairPainter != null) {
      _crosshairPainter.drawCrosshair(canvas);
    }
  }

  void _drawLine(Canvas canvas, Paint paint, int seriesIndex) {
    if (_crosshairPainter != null ) {
      _crosshairPainter.drawCrosshairLine(canvas, paint, seriesIndex);
    }
  }

  Paint _linePainter(Paint paint) {
    return _crosshairPainter?.getLinePainter(paint);
  }
}

class _ChartPointInfo {
  /// Marker x position
  double markerXPos;

  /// Marker y position
  double markerYPos;

  /// label for trackball and cross hair
  String label;

  /// Data point index
  int dataPointIndex;

  /// Instance of chart series
  XyDataSeries<dynamic, dynamic> series;

  /// Chart data point
  _CartesianChartPoint<dynamic> chartDataPoint;

  /// X position of the label
  double xPosition;

  /// Y position of the label
  double yPosition;

  /// Color of the segment
  Color color;
}
