part of charts;

class TrackballBehavior extends ChartBehavior {
  TrackballBehavior({
    ActivationMode activationMode,
    TrackballLineType lineType,
    TrackballDisplayMode tooltipDisplayMode,
    ChartAlignment tooltipAlignment,
    InteractiveTooltip tooltipSettings,
    this.lineDashArray,
    this.enable = false,
    this.lineColor,
    this.lineWidth = 1,
    this.shouldAlwaysShow = false,
  })
      : activationMode = activationMode ?? ActivationMode.longPress,
        tooltipAlignment =
            tooltipAlignment ?? ChartAlignment.center,
        tooltipDisplayMode = tooltipDisplayMode ?? TrackballDisplayMode.floatAllPoints,
        tooltipSettings = tooltipSettings ?? InteractiveTooltip(),
        lineType = lineType ?? TrackballLineType.vertical{
           _isLongPressActivated = false;
        }

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true),
  ///        ));
  ///}
  ///```
  final bool enable;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, lineWidth: 5),
  ///        ));
  ///}
  ///```
  final double lineWidth;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, lineColor: Colors.red),
  ///        ));
  ///}
  ///```
  final Color lineColor;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, lineDashArray: [10,10]),
  ///        ));
  ///}
  ///```
  final List<double> lineDashArray;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, activationMode: ActivationMode.doubleTap),
  ///        ));
  ///}
  ///```
  final ActivationMode activationMode;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, tooltipAlignment: ChartAlignment.far),
  ///        ));
  ///}
  ///```
  final ChartAlignment tooltipAlignment;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, lineType: TrackballLineType.horizontal),
  ///        ));
  ///}
  ///```
  final TrackballLineType lineType;

  final TrackballDisplayMode tooltipDisplayMode;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///           trackballBehavior: TrackballBehavior(enable: true, shouldAlwaysShow: true),
  ///        ));
  ///}
  ///```
  final bool shouldAlwaysShow;

  /// Setting model for tooltip.
  InteractiveTooltip tooltipSettings;

  /// Touch position
  Offset _position;

  /// Holds the instance of trackballPainter
  _TrackballPainter _trackballPainter;

  /// Check whether long press activated or not .
  bool _isLongPressActivated;

  void show(double x, double y) {
    if (_trackballPainter != null && activationMode!=ActivationMode.none && x !=null && y !=null) {
      _trackballPainter.generateAllPoints(Offset(x, y));
      _trackballPainter.canResetPath = false;
      _trackballPainter.chart._chartState.trackballRepaintNotifier.value++;
    }
  }

  void hide() {
    if (_trackballPainter != null && activationMode!=ActivationMode.none) {
      _trackballPainter.canResetPath = true;
      _trackballPainter.chart._chartState.trackballRepaintNotifier.value++;
    }
  }

  @override
  void onDoubleTap(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onLongPress(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onTouchDown(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onTouchMove(double xPos, double yPos) {
    show(xPos, yPos);
  }

  @override
  void onTouchUp(double xPos, double yPos) {
    hide();
  }

  @override
  void onPaint(Canvas canvas) {
    if (_trackballPainter != null && !_trackballPainter.canResetPath) {
      _trackballPainter.drawTrackball(canvas);
    }
  }

  void _drawLine(Canvas canvas, Paint paint, int seriesIndex) {
    if (_trackballPainter != null) {
      _trackballPainter.drawTrackBallLine(canvas, paint, seriesIndex);
    }
  }

  Paint _linePainter(Paint paint) {
    return _trackballPainter?.getLinePainter(paint);
  }
}
