part of charts;

typedef CircularLegendRenderCallback = LegendRenderArgs Function(
    LegendRenderArgs legendRenderArgs);
typedef CircularTooltipCallback = TooltipArgs Function(TooltipArgs tooltipArgs);
typedef CircularDatalabelRenderCallback = DataLabelRenderArgs Function(
    DataLabelRenderArgs dataLabelArgs);
typedef CircularPointTapCallback = PointTapArgs Function(
    PointTapArgs pointTapArgs);
typedef CircularSelectionCallback = SelectionArgs Function(
    SelectionArgs selectionArgs);

///ignore:must_be_immutable
class SfCircularChart extends StatefulWidget {
  SfCircularChart(
      {Key key,
        this.backgroundColor,
        this.backgroundImage,
        this.annotations,
        this.borderColor = Colors.transparent,
        this.borderWidth = 0.0,
        this.onLegendItemRender,
        this.onTooltipRender,
        this.dataLabelRender,
        this.onPointTapped,
        this.onLegendTapped,
        this.onSelectionChanged,
        this.palette = const <Color>[
          Color.fromRGBO(53,92,125,1),
          Color.fromRGBO(192,108,132,1),
          Color.fromRGBO(246,114,128,1),
          Color.fromRGBO(248,177,149,1),
          Color.fromRGBO(116,180,155,1),
          Color.fromRGBO(0,168,181,1),
          Color.fromRGBO(73,76,162,1),
          Color.fromRGBO(255,205,96,1),
          Color.fromRGBO(255,240,219,1),
          Color.fromRGBO(238,238,238,1)
        ],
        ChartTitle title,
        EdgeInsets margin,
        List<CircularSeries<dynamic, dynamic>> series,
        Legend legend,
        String centerX,
        String centerY,
        TooltipBehavior tooltipBehavior,
        List<IndexesModel> initialSelectedDataIndexes,
        ActivationMode selectionGesture,
        bool enableMultiSelection})
      : initialSelectedDataIndexes =
      initialSelectedDataIndexes ?? <IndexesModel>[],
        series = series = series ?? <CircularSeries<dynamic, dynamic>>[],
        title = title ?? ChartTitle(),
        legend = legend ?? Legend(),
        margin = margin ?? const EdgeInsets.fromLTRB(10, 10, 10, 10),
        centerX = centerX ?? '50%',
        centerY = centerY ?? '50%',
        tooltipBehavior = tooltipBehavior ?? TooltipBehavior(),
        selectionGesture = selectionGesture ?? ActivationMode.singleTap,
        enableMultiSelection = enableMultiSelection ?? false,
        super(key: key) {
    _chartSeries = _CircularSeries();
    _chartLegend = _ChartLegend();
    _chartTheme = _ChartTheme();
    _needToMoveFromCenter = true;
  }



  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            title: ChartTitle(text: 'Default rendering')
  ///        ));
  ///}
  ///```

  final ChartTitle title;



  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            series: <PieSeries<ChartData, String>>[
  ///              PieSeries<ChartData, String>(
  ///                enableTooltip: true,
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///```
  final List<CircularSeries<dynamic, dynamic>> series;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            margin: const EdgeInsets.all(2),
  ///            borderColor: Colors.blue
  ///        ));
  ///}
  ///```

  final EdgeInsets margin;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            legend: Legend(isVisible: true, position: LegendPosition.auto)
  ///        ));
  ///}
  ///```

  final Legend legend;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            tooltipBehavior: TooltipBehavior(enable: true)
  ///        ));
  ///}
  ///```

  final TooltipBehavior tooltipBehavior;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            backgroundColor: Colors.blue
  ///        ));
  ///}
  ///```

  final Color backgroundColor;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            annotations: <CircularChartAnnotation>[
  ///                CircularChartAnnotation(
  ///                   child Container(
  ///                   child: const Text('Empty data'),
  ///                    angle: 200,
  ///                    radius: '80%'
  ///                 ),
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///```
  final List<CircularChartAnnotation> annotations;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            borderColor: Colors.red
  ///        ));
  ///}
  ///```

  final Color borderColor;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            borderWidth: 2,
  ///            borderColor: Colors.red
  ///        ));
  ///}
  ///```

  final double borderWidth;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            initialSelectedDataIndexes: <IndexesModel>[IndexesModel(2,0)],
  ///           series: <PieSeries<ChartData, String>>[
  ///                PieSeries<ChartData, String>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final List<IndexesModel> initialSelectedDataIndexes;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            backgroundImage: const AssetImage('image.png'),
  ///        ));
  ///}
  ///```
  final ImageProvider backgroundImage;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            centerX: '50%'
  ///        ));
  ///}
  ///```
  final String centerX;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            centerY: '50%'
  ///        ));
  ///}
  ///```
  final String centerY;


  bool _needToMoveFromCenter;



  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            legend: Legend(isVisible: true),
  ///            onLegendItemRender: (LegendRendererArgs args) => legend(args),
  ///        ));
  ///}
  ///dynamic legend(LegendRendererArgs args) {
  ///   args.legendIconType = LegendIconType.diamond;
  ///}
  ///```
  final CircularLegendRenderCallback onLegendItemRender;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            tooltipBehavior: TooltipBehavior(enable: true)
  ///            onTooltipRender: (TooltipArgs args) => tool(args)
  ///        ));
  ///}
  ///dynamic tool(TooltipArgs args) {
  ///   args.locationX = 30;
  ///}
  ///```
  final CircularTooltipCallback onTooltipRender;



  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            dataLabelRender: (DataLabelRenderArgs args) => datalabel(args),
  ///            series: <PieSeries<ChartData, String>>[
  ///              PieSeries<ChartData, String>(
  ///                enableTooltip: true,
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///dynamic datalabel(DataLabelRenderArgs args) {
  ///    args.text = 'datalabel';
  ///}
  ///```
  final CircularDatalabelRenderCallback dataLabelRender;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            onPointTapped: (PointTapArgs args) => point(args),
  ///        ));
  ///}
  ///dynamic point(PointTapArgs args) {
  ///   print(args.seriesIndex);
  ///}
  ///```
  final CircularPointTapCallback onPointTapped;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            onLegendTapped: (LegendTapArgs args) => legend(args),
  ///        ));
  ///}
  ///dynamic legend(LegendTapArgs args) {
  ///   print(args.pointIndex);
  ///}
  ///```
  final ChartLegendTapCallback onLegendTapped;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            onSelectionChanged: (SelectionArgs args) => select(args),
  ///        ));
  ///}
  ///dynamic select(SelectionArgs args) {
  ///   print(args.selectedBorderColor);
  ///}
  ///```
  final CircularSelectionCallback onSelectionChanged;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            palette: <Color>[Colors.red, Colors.green]
  ///        ));
  ///}
  ///```
  final List<Color> palette;

  /// Holds the information of chart state class
  _SfCircularChartState _chartState;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            selectionGesture: ActivationMode.singleTap,
  ///            series: <PieSeries<ChartData, String>>[
  ///                PieSeries<ChartData, String>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```

  final ActivationMode selectionGesture;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCircularChart(
  ///            enableMultiSelection: true,
  ///            series: <PieSeries<ChartData, String>>[
  ///                PieSeries<ChartData, String>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```

  final bool enableMultiSelection;

  _ChartLegend _chartLegend;

  /// Chart theme for chart
  ///
  _ChartTheme _chartTheme;

  /// Holds the information of SeriesBase class
  _CircularSeries _chartSeries;

  @override
  State<StatefulWidget> createState() => _SfCircularChartState();
}

class _SfCircularChartState extends State<SfCircularChart>
    with TickerProviderStateMixin {
  List<AnimationController>
  controllerList; // Holds the animation controller list for all series
  AnimationController animationController; // Animation controller for series
  ValueNotifier<int>
  tooltipRepaintNotifier; // Repaint notifier for tooltip container
  AnimationController
  annotationController; // Animation controller for Annotations
  ValueNotifier<int>
  seriesRepaintNotifier; // Repaint notifier for series container
  List<_MeasureWidgetContext>
  legendWidgetContext; // To measure legend size and position
  List<_ChartTemplateInfo> templates; // Chart Template info
  List<Widget> _chartWidgets; // List of container widgets for chart series
  Offset centerLocation;
  Rect chartContainerRect;
  Rect chartAreaRect;
  bool animateCompleted;
  List<_Region> explodedRegions;
  List<_LegendRenderContext> legendToggleStates;
  List<_MeasureWidgetContext> legendToggleTemplateStates;
  bool initialRender;
  List<_ChartPoint<dynamic>> selectedDataPoints;
  List<_ChartPoint<dynamic>> unselectedDataPoints;
  List<_Region> selectedRegions;
  List<_Region> unselectedRegions;
  List<Rect> dataLabelTemplateRegions;
  _ChartTemplate _chartTemplate;

  @override
  void initState() {
    _initializeDefaultValues();
    super.initState();
  }

  void _initializeDefaultValues() {
    animateCompleted = false;
    initialRender = true;
    controllerList = <AnimationController>[];
    annotationController = AnimationController(vsync: this);
    tooltipRepaintNotifier = ValueNotifier<int>(0);
    seriesRepaintNotifier = ValueNotifier<int>(0);
    legendWidgetContext = <_MeasureWidgetContext>[];
    explodedRegions = <_Region>[];
    templates = <_ChartTemplateInfo>[];
    legendToggleStates = <_LegendRenderContext>[];
    selectedDataPoints = <_ChartPoint<dynamic>>[];
    unselectedDataPoints = <_ChartPoint<dynamic>>[];
    selectedRegions = <_Region>[];
    unselectedRegions = <_Region>[];
    legendToggleTemplateStates = <_MeasureWidgetContext>[];
    dataLabelTemplateRegions = <Rect>[];
  }

  @override
  void didUpdateWidget(SfCircularChart oldWidget) {
    _needsRepaintCircularChart(widget, oldWidget);
    oldWidget._chartState.animateCompleted = false;
    oldWidget._chartState.initialRender = false;
    super.didUpdateWidget(oldWidget);
  }

  @override
  void dispose() {
    if (controllerList != null) {
      for (int index = 0; index < controllerList.length; index++) {
        controllerList[index]
            .dispose(); // Need to dispose all the animation controller instance otherwise it will cause memory leak
      }
    }
    super.dispose();
  }

  void _redraw() {
    widget._chartState.initialRender = false;
    setState(() {});
  }

  void _refresh() {
    if (widget._chartState.legendWidgetContext.isNotEmpty) {
      for (int i = 0; i < widget._chartState.legendWidgetContext.length; i++) {
        final _MeasureWidgetContext templateContext =
        widget._chartState.legendWidgetContext[i];
        final RenderBox renderBox = templateContext.context.findRenderObject();
        templateContext.size = renderBox.size;
      }
      setState(() {});
    }
  }

  @override
  Widget build(BuildContext context) {
    final ThemeData theme = Theme.of(context);
    widget._chartTheme.initializeChartTheme(theme);
    widget._chartState = this;
    return Container(
      child: GestureDetector(
          child: Container(
            decoration: BoxDecoration(
                color: widget.backgroundColor ??
                    widget._chartTheme.plotAreaBackgroundColor,
                image: widget.backgroundImage != null
                    ? DecorationImage(
                    image: widget.backgroundImage, fit: BoxFit.fill)
                    : null,
                border: Border.all(
                    color: widget.borderColor, width: widget.borderWidth)),
            child: Column(
              children: <Widget>[_renderChartTitle(), _renderChartElements()],
            ),
          )),
    );
  }

  Widget _renderChartTitle() {
    Widget titleWidget;
    if (widget.title.text != null && widget.title.text.isNotEmpty) {
      final Color color =
          widget.title.textStyle.color ?? widget._chartTheme.chartTitleColor;
      final double fontSize = widget.title.textStyle.fontSize;
      final String fontFamily = widget.title.textStyle.fontFamily;
      final FontStyle fontStyle = widget.title.textStyle.fontStyle;
      final FontWeight fontWeight = widget.title.textStyle.fontWeight;
      titleWidget = Container(
        margin: EdgeInsets.fromLTRB(widget.margin.left, widget.margin.top,
            widget.margin.right, widget.margin.bottom),
        child: Container(
            padding: const EdgeInsets.fromLTRB(0, 5, 0, 0),
            decoration: BoxDecoration(
                color: widget.title.backgroundColor ??
                    widget._chartTheme.titleBackgroundColor,
                border: Border.all(
                    color: widget.title.borderColor ??
                        widget._chartTheme.chartTitleColor,
                    width: widget.title.borderWidth)),
            child: Text(widget.title.text,
                style: TextStyle(
                    color: color,
                    fontSize: fontSize,
                    fontFamily: fontFamily,
                    fontStyle: fontStyle,
                    fontWeight: fontWeight),
                textScaleFactor: 1.2,
                overflow: TextOverflow.clip,
                textAlign: TextAlign.center)),
        alignment: (widget.title.alignment == ChartAlignment.near)
            ? Alignment.topLeft
            : (widget.title.alignment == ChartAlignment.far)
            ? Alignment.topRight
            : (widget.title.alignment == ChartAlignment.center)
            ? Alignment.topCenter
            : Alignment.topCenter,
      );
    } else {
      titleWidget = Container();
    }
    return titleWidget;
  }

  Widget _renderChartElements() {
    return Expanded(
      child: LayoutBuilder(
          builder: (BuildContext context, BoxConstraints constraints) {
            Widget element;
            _initialize(constraints);
            widget._chartSeries._findVisibleSeries();
            widget._chartSeries._processDataPoints();
            final List<Widget> legendTemplates = _bindLegendTemplateWidgets();
            if (legendTemplates.isNotEmpty && legendWidgetContext.isEmpty) {
              element = Container(child: Stack(children: legendTemplates));
              SchedulerBinding.instance.addPostFrameCallback((_) => _refresh());
            } else {
              widget._chartLegend
                  ._calculateLegendBounds(widget._chartLegend.chartSize);
              element =
                  _getElements(widget, _CircularArea(chart: widget), constraints);
            }
            return element;
          }),
    );
  }

  void _initialize(BoxConstraints constraints) {
    widget._chartLegend.chart = widget;
    widget._chartSeries.chart = widget;
    final dynamic width = constraints.maxWidth;
    final dynamic height = constraints.maxHeight;
    final EdgeInsets margin = widget.margin;
    widget._chartLegend.chartSize = Size(width - margin.left - margin.right,
        height - margin.top - margin.bottom);
  }

  List<Widget> _bindLegendTemplateWidgets() {
    Widget legendWidget;
    final List<Widget> templates = <Widget>[];
    widget._chartState._chartWidgets = <Widget>[];
    if (widget.legend.isVisible && widget.legend.legendItemBuilder != null) {
      for (int i = 0; i < widget._chartSeries.visibleSeries.length; i++) {
        final CircularSeries<dynamic, dynamic> series =
        widget._chartSeries.visibleSeries[i];
        for (int j = 0; j < series._renderPoints.length; j++) {
          legendWidget = widget.legend.legendItemBuilder(
              series._renderPoints[j].x, series, series._renderPoints[j], j);
          templates.add(_MeasureWidgetSize(
              chart: widget,
              type: 'Legend',
              seriesIndex: i,
              pointIndex: j,
              currentKey: GlobalKey(),
              currentWidget: legendWidget,
              opacityValue: 0.0));
        }
      }
    }
    return templates;
  }
}

// ignore: must_be_immutable
class _CircularArea extends StatelessWidget {
  // ignore: prefer_const_constructors_in_immutables
  _CircularArea({this.chart});
  final SfCircularChart chart;
  CircularSeries<dynamic, dynamic> series;
  _ChartInteraction currentActive;
  Offset _tapPosition;

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
        builder: (BuildContext context, BoxConstraints constraints) {
          return Listener(
              onPointerUp: (PointerUpEvent event) =>
                  _onTapUp(context, event.position),
              onPointerDown: (PointerDownEvent event) =>
                  _onTapDown(context, event.position),
              child: GestureDetector(
                  onLongPress: onLongPress,
                  onDoubleTap: onDoubleTap,
                  child: Container(
                      decoration: const BoxDecoration(color: Colors.transparent),
                      child: _initializeChart(constraints, context))));
        });
  }

  RenderBox renderBox;
  _Region pointRegion;
  TapDownDetails tapDownDetails;
  Offset doubleTapPosition;
  void _onTapDown(BuildContext context, Offset globalPosition) {
    renderBox = context.findRenderObject();
    currentActive = null;
    _tapPosition = renderBox.globalToLocal(globalPosition);
    pointRegion = _getPointRegion(chart, _tapPosition);
    doubleTapPosition = _tapPosition;
    if (_tapPosition != null &&
        pointRegion != null &&
        chart.selectionGesture == ActivationMode.singleTap) {
      currentActive = _ChartInteraction(
          pointRegion.seriesIndex,
          pointRegion.pointIndex,
          chart._chartSeries.visibleSeries[pointRegion.seriesIndex],
          chart._chartSeries.visibleSeries[pointRegion.seriesIndex]
              ._renderPoints[pointRegion.pointIndex],
          pointRegion);
      series.selectionSettings._selectionRenderer.series = series;
      series.selectionSettings
          .onCircularTouchDown(pointRegion.pointIndex, pointRegion.seriesIndex);
    } else {
      chart.tooltipBehavior?._tooltipTemplate?.show = false;
      chart.tooltipBehavior?._tooltipTemplate?.state?._performTooltip();
    }
  }

  void onDoubleTap() {
    if (doubleTapPosition != null &&
        pointRegion != null &&
        chart.selectionGesture == ActivationMode.doubleTap) {
      currentActive = _ChartInteraction(
          pointRegion.seriesIndex,
          pointRegion.pointIndex,
          chart._chartSeries.visibleSeries[pointRegion.seriesIndex],
          chart._chartSeries.visibleSeries[pointRegion.seriesIndex]
              ._renderPoints[pointRegion.pointIndex],
          pointRegion);
      if (currentActive != null) {
        if (currentActive.series.explodeGesture == ActivationMode.doubleTap) {
          chart._chartSeries._seriesPointExplosion(currentActive.region);
        }
      }
      series.selectionSettings._selectionRenderer.series = series;
      series.selectionSettings
          .onCircularDoubleTap(pointRegion.pointIndex, pointRegion.seriesIndex);
      if (chart.tooltipBehavior.enable &&
          chart.tooltipBehavior.activationMode ==
              ActivationMode.doubleTap) {
        chart.tooltipBehavior.onDoubleTap(
            doubleTapPosition.dx.toDouble(), doubleTapPosition.dy.toDouble());
      }
    }
  }

  void onLongPress() {
    if (_tapPosition != null &&
        pointRegion != null &&
        chart.selectionGesture == ActivationMode.longPress) {
      currentActive = _ChartInteraction(
          pointRegion.seriesIndex,
          pointRegion.pointIndex,
          chart._chartSeries.visibleSeries[pointRegion.seriesIndex],
          chart._chartSeries.visibleSeries[pointRegion.seriesIndex]
              ._renderPoints[pointRegion.pointIndex],
          pointRegion);
      series.selectionSettings._selectionRenderer.series = series;
      series.selectionSettings
          .onCircularLongPress(pointRegion.pointIndex, pointRegion.seriesIndex);
      if (currentActive != null) {
        if (currentActive.series.explodeGesture == ActivationMode.longPress) {
          chart._chartSeries._seriesPointExplosion(currentActive.region);
        }
      }
      if (chart.tooltipBehavior.enable &&
          chart.tooltipBehavior.activationMode ==
              ActivationMode.longPress) {
        chart.tooltipBehavior.onLongPress(
            _tapPosition.dx.toDouble(), _tapPosition.dy.toDouble());
      }
    }
  }

  void _onTapUp(BuildContext context, Offset globalPosition) {
    if (_tapPosition != null) {
      if (currentActive != null &&
          currentActive.series != null &&
          currentActive.series.explodeGesture == ActivationMode.singleTap) {
        chart._chartSeries._seriesPointExplosion(currentActive.region);
      }
      if (chart.tooltipBehavior.enable &&
          chart.tooltipBehavior.activationMode ==
              ActivationMode.singleTap &&
          currentActive != null &&
          currentActive.series != null) {
        if (chart.tooltipBehavior.builder != null) {
          final CircularSeries<dynamic, dynamic> chartSeries =
              currentActive.series;
          final _ChartPoint<dynamic> point = currentActive.point;
          final Offset location = _degreeToPoint(point.midAngle,
              (point.innerRadius + point.outerRadius) / 2, point.center);
          chart.tooltipBehavior._tooltipTemplate.rect =
              Rect.fromLTWH(location.dx, location.dy, 0, 0);
          chart.tooltipBehavior._tooltipTemplate.template =
              chart.tooltipBehavior.builder(
                  chartSeries.dataSource[currentActive.pointIndex],
                  point,
                  chartSeries,
                  currentActive.seriesIndex,
                  currentActive.pointIndex);
          chart.tooltipBehavior._tooltipTemplate.show = true;
          chart.tooltipBehavior._tooltipTemplate?.state?._performTooltip();
        } else {
          final RenderBox renderBox = context.findRenderObject();
          final Offset position = renderBox.globalToLocal(globalPosition);
          chart.tooltipBehavior
              .onTouchUp(position.dx.toDouble(), position.dy.toDouble());
        }
      }
    }
    _tapPosition = null;
  }

  Widget _initializeChart(BoxConstraints constraints, BuildContext context) {
    _calculateSize(constraints);
    chart._chartSeries?._calculateAngleAndCenterPositions();
    return GestureDetector(
        child: Container(
            decoration: const BoxDecoration(color: Colors.transparent),
            child: _renderWidgets(constraints, context)));
  }

  void _calculateSize(BoxConstraints constraints) {
    final dynamic width = constraints.maxWidth;
    final dynamic height = constraints.maxHeight;
    chart._chartState.chartContainerRect = Rect.fromLTWH(0, 0, width, height);
    final EdgeInsets margin = chart.margin;
    chart._chartState.chartAreaRect = Rect.fromLTWH(
        margin.left,
        margin.top,
        width - margin.right - margin.left,
        height - margin.top - margin.bottom);
  }

  Widget _renderWidgets(BoxConstraints constraints, BuildContext context) {
    _bindSeriesWidgets();
    _findTemplates();
    _renderTemplates();
    _bindTooltipWidgets(constraints);
    return Container(child: Stack(children: chart._chartState._chartWidgets));
  }

  void _findTemplates() {
    Offset labelLocation;
    const num lineLength = 10;
    _ChartPoint<dynamic> point;
    Widget labelWidget;
    chart._chartState.templates = <_ChartTemplateInfo>[];
    chart._chartState.dataLabelTemplateRegions = <Rect>[];
    for (int k = 0; k < chart._chartSeries.visibleSeries.length; k++) {
      final CircularSeries<dynamic, dynamic> series =
      chart._chartSeries.visibleSeries[k];
      final ConnectorLineSettings connector =
          series.dataLabelSettings.connectorLineSettings;
      if (series.dataLabelSettings.isVisible && series.dataLabelSettings.builder != null) {
        for (int i = 0; i < series._renderPoints.length; i++) {
          point = series._renderPoints[i];
          ChartAlignment labelAlign;
          if (point.isVisible) {
            labelWidget = series.dataLabelSettings
                .builder(series.dataSource[i], point, series, i, k);
            if (series.dataLabelSettings.labelPosition == LabelPosition.inside) {
              labelLocation = _degreeToPoint(point.midAngle,
                  (point.innerRadius + point.outerRadius) / 2, point.center);
              labelLocation = Offset(labelLocation.dx, labelLocation.dy);
              labelAlign = ChartAlignment.center;
            } else {
              final num connectorLength =
              _percentToValue(connector.length, point.outerRadius);
              labelLocation = _degreeToPoint(point.midAngle,
                  point.outerRadius + connectorLength, point.center);
              labelLocation = Offset(
                  point.dataLabelPosition == Position.right
                      ? labelLocation.dx + lineLength + 5
                      : labelLocation.dx - lineLength - 5,
                  labelLocation.dy);
              labelAlign = point.dataLabelPosition == Position.left
                  ? ChartAlignment.far
                  : ChartAlignment.near;
            }
            chart._chartState.templates.add(_ChartTemplateInfo(
                key: GlobalKey(),
                templateType: 'DataLabel',
                pointIndex: i,
                seriesIndex: k,
                needMeasure: true,
                clipRect: chart._chartState.chartAreaRect,
                animationDuration: 500,
                widget: labelWidget,
                horizontalAlignment: labelAlign,
                verticalAlignment: ChartAlignment.center,
                location: labelLocation));
          }
        }
      }
    }

    if (chart.annotations != null && chart.annotations.isNotEmpty) {
      for (int i = 0; i < chart.annotations.length; i++) {
        final CircularChartAnnotation annotation = chart.annotations[i];
        if (annotation.child != null) {
          final double radius =
          _percentToValue(annotation.radius, chart._chartSeries.size / 2);
          final Offset point = _degreeToPoint(
              annotation.angle, radius, chart._chartState.centerLocation);
          final double annotationHeight =
          _percentToValue(annotation.height, chart._chartSeries.size / 2);
          final double annotationWidth =
          _percentToValue(annotation.width, chart._chartSeries.size / 2);
          final _ChartTemplateInfo templateInfo = _ChartTemplateInfo(
              key: GlobalKey(),
              templateType: 'Annotation',
              needMeasure: true,
              horizontalAlignment: annotation.horizontalAlignment,
              verticalAlignment: annotation.verticalAlignment,
              clipRect: chart._chartState.chartContainerRect,
              widget: annotationHeight > 0 && annotationWidth > 0
                  ? Container(
                  height: annotationHeight,
                  width: annotationWidth,
                  child: annotation.child)
                  : annotation.child,
              pointIndex: i,
              animationDuration: 500,
              location: point);

          chart._chartState.templates.add(templateInfo);
        }
      }
    }
  }

  void _renderTemplates() {
    if (chart._chartState.templates.isNotEmpty) {
      for (int i = 0; i < chart._chartState.templates.length; i++) {
        final _ChartTemplateInfo chartTemplateInfo =
        chart._chartState.templates[i];
        chartTemplateInfo.animationDuration = !chart._chartState.initialRender
            ? 0
            : chartTemplateInfo.animationDuration;
      }
      chart._chartState._chartTemplate = _ChartTemplate(
          templates: chart._chartState.templates,
          render: chart._chartState.animateCompleted,
          chart: chart);
      chart._chartState._chartWidgets.add(chart._chartState._chartTemplate);
    }
  }

  void _bindTooltipWidgets(BoxConstraints constraints) {
    if (chart.tooltipBehavior.enable) {
      if (chart.tooltipBehavior.builder != null) {
        chart.tooltipBehavior._tooltipTemplate = _TooltipTemplate(
            show: false, clipRect: chart._chartState.chartContainerRect);
        chart._chartState._chartWidgets
            .add(chart.tooltipBehavior._tooltipTemplate);
      } else {
        final _TooltipPainter tooltipPainter = _TooltipPainter(
            valueNotifier: chart._chartState.tooltipRepaintNotifier);
        tooltipPainter._chartState = chart._chartState;
        tooltipPainter._chart = chart;
        tooltipPainter.tooltip = chart.tooltipBehavior;
        chart.tooltipBehavior._painter = tooltipPainter;
        chart._chartState._chartWidgets.add(
          Container(
              height: constraints.maxHeight,
              width: constraints.maxWidth,
              decoration: const BoxDecoration(color: Colors.transparent),
              child: CustomPaint(painter: tooltipPainter)),
        );
      }
    }
  }

  void _bindSeriesWidgets() {
    CustomPainter seriesPainter;
    Animation<double> seriesAnimation;
    for (int i = 0; i < chart._chartSeries.visibleSeries.length; i++) {
      series = chart._chartSeries.visibleSeries[i];
      if (series.selectionSettings != null) {
        series.selectionSettings._selectionRenderer ??= _SelectionRenderer();
        series.selectionSettings._selectionRenderer.chart = chart;
        series.selectionSettings._selectionRenderer.selectedDataPoints =
            chart._chartState.selectedDataPoints;
        series.selectionSettings._selectionRenderer.unselectedDataPoints =
            chart._chartState.unselectedDataPoints;
        series.selectionSettings._selectionRenderer.selectedRegions =
            chart._chartState.selectedRegions;
        series.selectionSettings._selectionRenderer.unselectedRegions =
            chart._chartState.unselectedRegions;
        if (chart.initialSelectedDataIndexes.isNotEmpty) {
          series.selectionSettings._selectionRenderer.chart = chart;
          series.selectionSettings._selectionRenderer.series = series;
          series.selectionSettings
              ._selectedDataPointIndex(chart.initialSelectedDataIndexes);
        }
      }
      if (series.animationDuration > 0 && chart._chartState.initialRender) {
        chart._chartState.animateCompleted = false;
        chart._chartState.animationController =
        AnimationController(vsync: chart._chartState)
          ..addListener(() {
            chart._chartState.seriesRepaintNotifier.value++;
          });
        chart._chartState.animationController.duration =
            Duration(milliseconds: series.animationDuration.toInt());
        seriesAnimation =
            Tween<double>(begin: 0.0, end: 1.0).animate(CurvedAnimation(
              parent: chart._chartState.animationController,
              curve: const Interval(0.1, 0.8, curve: Curves.linear),
            )..addStatusListener((AnimationStatus status) {
              if (status == AnimationStatus.completed) {
                if (chart._chartState._chartTemplate != null &&
                    chart._chartState._chartTemplate.state != null) {
                  chart._chartState._chartTemplate.state.templateRender();
                }
              }
            }));
        chart._chartState.animationController.forward(from: 0.0);
        chart._chartState.controllerList
            .add(chart._chartState.animationController);
      } else {
        chart._chartState.animateCompleted = true;
      }
      _animateComplete(chart._chartState.controllerList, chart);
      if (series._seriesType == 'pie') {
        seriesPainter = _PieChartPainter(
            chart: chart,
            index: i,
            isRepaint: series._needsRepaint,
            animationController: chart._chartState.animationController,
            seriesAnimation: seriesAnimation,
            notifier: chart._chartState.seriesRepaintNotifier);
      } else if (series._seriesType == 'doughnut') {
        seriesPainter = _DoughnutChartPainter(
            chart: chart,
            index: i,
            isRepaint: series._needsRepaint,
            animationController: chart._chartState.animationController,
            seriesAnimation: seriesAnimation,
            notifier: chart._chartState.seriesRepaintNotifier);
      } else if (series._seriesType == 'radialbar') {
        seriesPainter = _RadialBarPainter(
            chart: chart,
            index: i,
            isRepaint: series._needsRepaint,
            animationController: chart._chartState.animationController,
            seriesAnimation: seriesAnimation,
            notifier: chart._chartState.seriesRepaintNotifier);
      }
      chart._chartState._chartWidgets
          .add(RepaintBoundary(child: CustomPaint(painter: seriesPainter)));
    }
  }
}