part of charts;

class _ChartPoint<D> {
  _ChartPoint(this.x, this.y);
  dynamic x;
  num y;
  num degree;
  num startAngle;
  num endAngle;
  num midAngle;
  Offset center;
  String text;
  Color fill;
  Color color;
  Color strokeColor;
  D sortValue;
  num strokeWidth;
  num innerRadius;
  num outerRadius;
  bool isExplode;
  bool isShadow;
  bool isEmpty = false;
  bool isVisible = true;
  bool isSelected = false;
  Position dataLabelPosition;
  LabelPosition renderPosition;
  Rect labelRect;
  Size dataLabelSize = const Size(0, 0);
  bool saturationRegionOutside = false;
}

class _Region {
  _Region(this.start, this.end, this.startAngle, this.endAngle, this.seriesIndex,
      this.pointIndex, this.center, this.innerRadius, this.outerRadius);
  int seriesIndex;
  int pointIndex;
  num startAngle;
  num start;
  num end;
  num endAngle;
  Offset center;
  num innerRadius;
  num outerRadius;
}

class _StyleOptions {
  _StyleOptions(this.fill, this.strokeWidth, this.strokeColor, [this.opacity]);
  Color fill;
  Color strokeColor;
  double opacity;
  num strokeWidth;
}

class ConnectorLineSettings {
  ConnectorLineSettings(
      {String length, double width, ConnectorType type, this.color})
      : length = length ?? '10%',
        width = width ?? 1.0,
        type = type ?? ConnectorType.line;

  final String length;

  final double width;

  final Color color;

  final ConnectorType type;
}

class _ChartInteraction {
  _ChartInteraction(
      this.seriesIndex, this.pointIndex, this.series, this.point, this.region);
  int seriesIndex;
  int pointIndex;
  CircularSeries<dynamic, dynamic> series;
  _ChartPoint<dynamic> point;
  _Region region;
}

///To get circular series data label saturation color
Color _getCircularDatalabelColor(_ChartPoint<dynamic> currentPoint,
    CircularSeries<dynamic, dynamic> series, SfCircularChart chart) {
  Color color;
  final DataLabelSettings dataLabel = series.dataLabelSettings;
  final String _seriesType = series._seriesType == 'pie'
      ? 'Pie'
      : series._seriesType == 'doughnut'
          ? 'Doughnut'
          : series._seriesType == 'radialbar'
              ? 'RadialBar'
              : 'Default';
  switch (_seriesType) {
    case 'Pie':
    case 'Doughnut':
      if (currentPoint.renderPosition == LabelPosition.inside &&
          !currentPoint.saturationRegionOutside) {
        color = _getInnerColor(dataLabel.color, currentPoint.fill, chart._chartTheme);
      } else {
        color = _getOuterColor(dataLabel.color,
            chart.backgroundColor ?? chart._chartTheme.plotAreaBackgroundColor, chart._chartTheme);
      }
      break;
    case 'RadialBar':
      final RadialBarSeries<dynamic, dynamic> radialBar = series;
      color = radialBar.trackColor;
      break;
    default:
      color = Colors.white;
  }
  return _getSaturationColor(color);
}

///To get inner data label color
Color _getInnerColor(Color dataLabelColor, Color pointColor, _ChartTheme theme) =>
    dataLabelColor != null
        ? dataLabelColor
        : pointColor != null ? pointColor : theme.brightness == Brightness.light ? const Color.fromRGBO(255, 255, 255, 1) : Colors.black;

///To get outer data label color
Color _getOuterColor(Color dataLabelColor, Color backgroundColor, _ChartTheme theme) =>
    dataLabelColor != null
        ? dataLabelColor
        : backgroundColor != Colors.transparent
            ? backgroundColor
            : theme.brightness == Brightness.light ? const Color.fromRGBO(255, 255, 255, 1) : Colors.black;

class CircularChartAnnotation {
  CircularChartAnnotation(
      {int angle,
      String radius,
      this.child,
      String height,
      String width,
      ChartAlignment horizontalAlignment,
      ChartAlignment verticalAlignment})
      : angle = angle ?? 0,
        radius = radius ?? '0%',
        height = height ?? '0%',
        width = width ?? '0%',
        verticalAlignment = verticalAlignment ?? ChartAlignment.center,
        horizontalAlignment = horizontalAlignment ?? ChartAlignment.center;
  final int angle;
  final String radius;
  final Widget child;
  final String height;
  final String width;
  final ChartAlignment horizontalAlignment;
  final ChartAlignment verticalAlignment;
}

bool _checkIsAnyPointSelect(CircularSeries<dynamic, dynamic> series,
    _ChartPoint<dynamic> point, SfCircularChart chart) {
  bool isAnyPointSelected = false;
  if (chart.initialSelectedDataIndexes.isNotEmpty) {
    for (int i = 0; i < chart.initialSelectedDataIndexes.length; i++) {
      final IndexesModel indexesModel = chart.initialSelectedDataIndexes[i];
      for (int j = 0; j < series._renderPoints.length; j++) {
        if (j == indexesModel.pointIndex) {
          isAnyPointSelected = true;
          break;
        }
      }
    }
  }
  return isAnyPointSelected;
}
