part of charts;

class _ChartContainer extends SingleChildRenderObjectWidget {
  const _ChartContainer({Widget child}) : super(child: child);

  @override
  RenderObject createRenderObject(BuildContext context) {
    return _ChartContainerBox();
  }
}

class _ChartContainerBox extends RenderShiftedBox {
  _ChartContainerBox() : super(null);
  double minHeight = 300;
  double minWidth = 300;

  @override
  void performLayout() {
    double height = constraints.maxHeight;
    double width = constraints.maxWidth;
    if (height == double.infinity) {
      height = minHeight;
    }
    if (width == double.infinity) {
      width = minWidth;
    }
    child.layout(
        BoxConstraints(
          minHeight: 0.0,
          maxHeight: height,
          minWidth: 0.0,
          maxWidth: width,
        ),
        parentUsesSize: true);
    size = Size(
        width,
        height); // constraints.maxHeight become infinity when widget is placed inside row/column
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    super.paint(context, offset);
  }
}

class ChartTitle {
  ChartTitle(
      {this.text = '',
      ChartTextStyle textStyle,
      this.alignment = ChartAlignment.center,
      this.borderColor = Colors.transparent,
      this.borderWidth = 0,
      this.backgroundColor})
      : textStyle = _getTextStyle(
            textStyle: textStyle,
            fontSize: 15.0,
            fontFamily: 'Segoe UI',
            fontStyle: FontStyle.normal,
            fontWeight: FontWeight.normal);

  final String text;

  final ChartTextStyle textStyle;

  final ChartAlignment alignment;

  final Color backgroundColor;

  final Color borderColor;

  final double borderWidth;
}

class ChartTextStyle {
  ChartTextStyle(
      {this.color,
      this.fontFamily ='Roboto',
      this.fontStyle = FontStyle.normal,
      this.fontWeight = FontWeight.normal,
      this.fontSize = 12});
  Color color;
  String fontFamily;
  FontStyle fontStyle;
  FontWeight fontWeight;
  double fontSize;
}

class Legend {
  Legend(
      {bool isVisible,
      LegendPosition position,
      ChartAlignment alignment,
      this.backgroundColor,
      Color borderColor,
      double borderWidth,
      double opacity,
      this.height,
      this.width,
      double padding,
      double iconHeight,
      double iconWidth,
      bool toggleSeriesVisibility,
      ChartTextStyle textStyle,
      bool isResponsive,
      LegendItemOrientation orientation,
      LegendTitle title,
      LegendItemOverflowMode overflowMode,
      this.legendItemBuilder,
      Color iconBorderColor,
      double iconBorderWidth,
      double itemPadding})
      : isVisible = isVisible ?? false,
        position = position ?? LegendPosition.auto,
        alignment = alignment ?? ChartAlignment.center,
        borderColor = borderColor ?? Colors.transparent,
        borderWidth = borderWidth ?? 0.0,
        iconBorderColor = iconBorderColor ?? Colors.transparent,
        iconBorderWidth = iconBorderWidth ?? 0.0,
        opacity = opacity ?? 1.0,
        padding = padding ?? 5.0,
        textStyle = _getTextStyle(
            textStyle: textStyle,
            fontSize: 13.0,
            fontStyle: FontStyle.normal,
            fontFamily: 'Segoe UI'),
        iconHeight = iconHeight ?? 12.0,
        iconWidth = iconWidth ?? 12.0,
        toggleSeriesVisibility = toggleSeriesVisibility ?? true,
        isResponsive = isResponsive ?? false,
        orientation = orientation ?? LegendItemOrientation.auto,
        overflowMode = overflowMode ?? LegendItemOverflowMode.scroll,
        itemPadding = itemPadding ?? 10.0,
        title = title ?? LegendTitle(){
        _renderer = _LegendRenderer();
  }

  final bool isVisible;

  final LegendPosition position;

  final ChartAlignment alignment;

  final Color backgroundColor;

  final Color borderColor;

  final double borderWidth;

  final Color iconBorderColor;

  final double iconBorderWidth;

  final double opacity;

  final String height;

  final String width;

  final double padding;

  final double iconHeight;

  final double iconWidth;

  final bool toggleSeriesVisibility;

  final ChartTextStyle textStyle;

  final bool isResponsive;

  final LegendItemOrientation orientation;

  final LegendTitle title;

  final LegendItemBuilder legendItemBuilder;

  final LegendItemOverflowMode overflowMode;

  final double itemPadding;

  _LegendRenderer _renderer;

  LegendPosition legendPosition;

  LegendItemOrientation _orientation;
}

class _MeasureWidgetContext {
  _MeasureWidgetContext(
      {this.context, this.key, this.widget, this.seriesIndex, this.pointIndex});
  BuildContext context;
  int seriesIndex;
  int pointIndex;
  Key key;
  Size size;
  Widget widget;
  bool isRender = false;
}

class LegendTitle {
  LegendTitle({this.text, ChartTextStyle textStyle, ChartAlignment alignment})
      : textStyle = _getTextStyle(
            textStyle: textStyle,
            fontSize: 14.0,
            fontStyle: FontStyle.normal,
            fontFamily: 'Segoe UI'),
        alignment = alignment ?? ChartAlignment.center;
  final String text;
  final ChartTextStyle textStyle;
  final ChartAlignment alignment;
}

typedef LegendItemBuilder = Widget Function(
    String legendText, dynamic series, dynamic point, int seriesIndex);

class EmptyPointSettings {
  EmptyPointSettings(
      {this.color = Colors.black,
      this.mode = EmptyPointMode.gap,
      this.borderColor = Colors.blue,
      this.borderWidth = 2});
  final Color color;
  final Color borderColor;
  final double borderWidth;
  final EmptyPointMode mode;
}

typedef ChartIndexedValueMapper<R> = R Function(int index);
typedef ChartValueMapper<T, R> = R Function(T datum, int index);

typedef ChartWidgetBuilder<T> = Widget Function(dynamic data, dynamic point,
    dynamic series, int pointIndex, int seriesIndex);
