part of charts;

class ChartSeries<T, D> {
  ChartSeries({
    this.xValueMapper,
    this.yValueMapper,
    this.dataLabelMapper,
    this.name,
    this.dataSource,
    this.pointColorMapper,
    this.sortFieldValueMapper,
    this.enableTooltip,
    this.emptyPointSettings,
    this.dataLabelSettings,
    this.animationDuration,
    this.borderColor,
    this.borderWidth,
    this.selectionSettings,
    this.legendIconType,
    this.opacity,
    this.sortingOrder}){
      _needsRepaint = true;
    }

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  dataSource: chartData,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///final dynamic chartData = <SalesData>[
  ///    SalesData(DateTime(2005, 0, 1), 'India', 0.5, 16, 10, 680, 760),
  ///    SalesData(DateTime(2006, 0, 1), 'China', 1, 12, 7, 550, 880),
  ///    SalesData(DateTime(2007, 0, 1), 'USA', 1.5, 18, 11, 440, 788),
  ///    SalesData(DateTime(2008, 0, 1), 'Japan', 2, 14, 7, 350, 560),
  ///    SalesData(DateTime(2009, 0, 1), 'Russia', 2.5, 11, 8, 444, 566),
  ///    SalesData(DateTime(2010, 0, 1), 'France', 3, 14, 6, 780, 650),
  ///    SalesData(DateTime(2011, 0, 1), 'Germany', 3.5, 10, 3, 450, 800),
  ///    SalesData(DateTime(2012, 0, 1), 'Italy', 4, 16, 7, 450, 800)
  ///  ];
  ///```
  final List<T> dataSource;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  dataSource: chartData,
  ///                  xValueMapper: (SalesData sales, _) => sales.numeric,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///class SalesData {
  ///   SalesData(this.year, this.category, this.numeric, this.sales1, this.sales2, this.sales3, this.sales4);
  ///   final DateTime year;
  ///   final String category;
  ///   final double numeric;
  ///   final int sales1;
  ///   final int sales2;
  ///   final int sales3;
  ///   final int sales4;
  ///}
  ///```
  final ChartIndexedValueMapper<D> xValueMapper;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  dataSource: chartData,
  ///                  yValueMapper: (SalesData sales, _) => sales.sales3,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///class SalesData {
  ///   SalesData(this.year, this.category, this.numeric, this.sales1, this.sales2, this.sales3, this.sales4);
  ///   final DateTime year;
  ///   final String category;
  ///   final double numeric;
  ///   final int sales1;
  ///   final int sales2;
  ///   final int sales3;
  ///   final int sales4;
  ///}
  ///```
  final ChartIndexedValueMapper<num> yValueMapper;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataSource: chartData,
  ///                       pointColorMapper: (BubbleColors sales, _) => sales.pointColorMapper,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///final dynamic chartData = <BubbleColors>[
  ///    BubbleColors(92.2, 7.8, 1.347, const Color.fromRGBO(0, 0, 255, 1)),
  ///    BubbleColors(74, 6.5, 1.241, const Color.fromRGBO(255, 0, 0, 1)),
  ///    BubbleColors(90.4, 6.0, 0.238, const Color.fromRGBO(255, 100, 102, 1)),
  ///    BubbleColors(99.4, 2.2, 0.197, const Color.fromRGBO(122, 100, 255, 1)),
  /// ];
  ///class BubbleColors {
  ///  BubbleColors(this.year, this.growth,[this.bubbleSize, this.pointColorMapper]);
  ///  final num year;
  ///  final num growth;
  ///  final num bubbleSize;
  ///  final Color pointColorMapper;
  ///}
  ///```
  final ChartIndexedValueMapper<Color> pointColorMapper;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                 dataSource: <SalesData>[
  ///                    SalesData(DateTime(2005, 0, 1), 'India', Colors.red, 16, 10, 680, 760),
  ///                    SalesData(DateTime(2006, 0, 1), 'China', Colors.blue, 12, 7, 550, 880),
  ///                    SalesData(DateTime(2007, 0, 1), 'USA', Colors.green, 18, 11, 440, 788),
  ///                    SalesData(DateTime(2008, 0, 1), 'Japan', Colors.yellow, 14, 7, 350, 560),
  ///                ],
  ///                dataLabelMapper: (ChartData data, _) => data.category,
  ///              ),
  ///             ],
  ///        ));
  ///}
  ///class SalesData {
  ///   SalesData(this.year, this.category, this.pointColor, this.sales1, this.sales2, this.sales3, this.sales4);
  ///   final DateTime year;
  ///   final String category;
  ///   final Color pointColor;
  ///   final int sales1;
  ///   final int sales2;
  ///   final int sales3;
  ///   final int sales4;
  ///}
  ///```
  final ChartIndexedValueMapper<String> dataLabelMapper;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       emptyPointSettings: EmptyPointSettings(color: Colors.black),
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final EmptyPointSettings emptyPointSettings;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       dataLabelSettings: DataLabelSettings(isVisible: true),
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final DataLabelSettings dataLabelSettings;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       name: 'Bubble Series',
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final String name;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BubbleSeries<BubbleColors, num>>[
  ///                   BubbleSeries<BubbleColors, num>(
  ///                       enableTooltip: true,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final bool enableTooltip;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       animationDuration: 1500,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final double animationDuration;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       borderColor: Colors.red,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final Color borderColor;

 
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       borderWidth: 5,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final double borderWidth;

  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <LineSeries<SalesData, String>>[
  ///                   LineSeries<SalesData, String>(
  ///                       legendIconType: LegendIconType.diamond,
  ///                  ),
  ///             ],
  ///        ));
  ///}
  ///```
  final LegendIconType legendIconType;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                  selectionSettings: SelectionSettings(
  ///                    selectedColor: Colors.red.withOpacity(0.8),
  ///                    unselectedColor: Colors.grey.withOpacity(0.5)
  ///                  ),
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final SelectionSettings selectionSettings;


  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                      opacity: 1
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final double opacity;


  /// This property holds the sorting field in dataSource
  final ChartIndexedValueMapper<dynamic> sortFieldValueMapper;

  
  ///```dart
  ///Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfCartesianChart(
  ///            series: <BarSeries<SalesData, num>>[
  ///                BarSeries<SalesData, num>(
  ///                    sortingOrder: SortingOrder.ascending,
  ///                ),
  ///              ],
  ///        ));
  ///}
  ///```
  final SortingOrder sortingOrder;


  /// Name of the series
  String _seriesName;

  bool _visible;

  /// It specifies the show or hide the series.
  bool isVisible;

  /// Whether to repaint series or not
  bool _needsRepaint;

}

